#run with: python3 -m unittest TestDecisionTreeModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        self.dofull = ( "DOFULL" in os.environ and os.environ["DOFULL"] == "1" )

    def test1( self ):
        print( "  Test 1 DecisionTreeModel" )
        dtm = DecisionTreeModel( )


        print( dtm )
        print( dtm.npars )
        cdtm = dtm.copy()
        print( cdtm )
        print( cdtm.npars )

        dtm1 = DecisionTreeModel( ndim=9, kdim=[0], depth=1 )
        print( dtm1 )
        print( dtm1.npars )

        dtm2 = DecisionTreeModel( ndim=9, kdim=[0,1,2], depth=2 )
        print( dtm2 )
        print( dtm2.npars )

        dtm3 = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6],
                    split=[0.1*k for k in range(2,8)], depth=3 )
        print( dtm3 )
        print( dtm3.npars )
        cdtm = dtm3.copy()
        print( cdtm )
        print( cdtm.npars )

        print( "============================" )
        wk = dtm3.walk()
        while True :
            try :
                pm = next( wk )
                if pm is None :
                    print( "TEST0  ", pm )
                else :
                    print( "TEST1  ", pm.shortName(), pm.dimension )
            except :
                break

        print( dtm3.findLeaf( 0 ) )
        print( dtm3.findLeaf( 1 ) )
        print( dtm3.findLeaf( 2 ) )
        print( dtm3.findLeaf( 4 ) )
        print( dtm3.findLeaf( 6 ) )

        printclass( dtm3 )
        printclass( cdtm )

        printclass( dtm3.left )
        printclass( cdtm.left )

    def test1a( self ):
        print( "  Test 1a DecisionTreeModel" )
        dtm = DecisionTreeModel( )
        print( dtm, dtm.isModifiable(), dtm.isDynamic() )
        cdtm = dtm.copy()
        print( cdtm, cdtm.isModifiable(), cdtm.isDynamic() )

        self.assertTrue( dtm.isModifiable() )
        self.assertTrue( cdtm.isModifiable() )

        dtm = DecisionTreeModel( ndim=9, kdim=[0], depth=1, modifiable=False )
        print( dtm, dtm.isModifiable(), dtm.modifiable )
        cdtm = dtm.copy()
        print( cdtm, cdtm.isModifiable(), cdtm.modifiable )

        self.assertFalse( dtm.isModifiable() )
        self.assertFalse( cdtm.isModifiable() )

    def test2( self ):
        print( "  Test 2 DecisionTreeModel" )
        ND = 10
        NP = 100
        m = DecisionTreeModel( ndim=ND, kdim=[0,1,2,3,4,5,6], depth=3 )
        print( m )

        numpy.random.seed( 12345 )
        xdata = numpy.random.rand( NP, ND )
        print( fmt( xdata ) )
        par0 = numpy.arange( m.npars, dtype=float ) + 1

        res = m.result( xdata, par0 )
        part = m.partial( xdata, par0 )

        for k in range( NP ) :
            print( fmt( res[k] ), fmt( part[k,:], max=None ) )
            self.assertTrue( part[k,int(res[k]-0.99)] == 1 )
            self.assertTrue( numpy.sum( part[k,:] ) == 1 )


    def test3( self ):
        print( "  Test 3 DecisionTreeModel" )

        for k in range( 4 ) :
            print( "==== depth : %d ============" % k )
            dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6,7,8,9],
                    split=0.3+0.1*k, itypes=[0,1,3,7]*5, depth=k )
            print( dtm.fullName() )

            print( "CODE  ", dtm.encode() )

        dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6,7,8,9],
                    split=0.5, depth=0, itypes=[0,1,3,5]*5 )
        for k in range( 4 ) :
            print( "==== grow  : %d ============" % k )
            dtm.grow( location=0, kdim=2+k, split=0.2+0.1*k )
            print( dtm.fullName() )
            print( "CODE  ", dtm.encode() )

#        dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6,7,8,9],
#                    split=0.5, depth=0, itypes=[0,1,3,5]*5 )
        for k in range( 4 ) :
            print( "==== grow  : %d ============" % k )
            loc = dtm.npars - 1
            dtm.grow( location=loc, kdim=2+k, split=0.2+0.1*k )
            print( dtm.fullName() )
            print( "CODE  ", dtm.encode() )

        code, dims, splim = dtm.encode()
        dt2 = DecisionTreeModel( code=code, kdim=dims, split=splim )
        print( "CODE  ", dt2.encode() )
        print( dt2.fullName( ids=False ) )



    def test4( self ):
        print( "  Test 4 DecisionTreeModel: rencode" )

        for dep in range( 4 ) :
            dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6,7,8,9],
                    split=0.5, depth=dep, itypes=[0,1,3,5]*5 )
            code, dims, splim = dtm.encode()
            print( "DEPTH  ", dep, " CODE  ", code, dims, splim )
            print( dtm.fullName( ids=False ) )


            dt2 = DecisionTreeModel( code=code, kdim=dims, split=splim )
            print( "CODE  ", dt2.encode() )
            print( dt2.fullName( ids=False ) )

        dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6,7,8,9],
                    split=0.5, depth=0, itypes=[0,1,3,5]*5 )
        dtm.grow( location=0, kdim=2, split=0.2 )
        dtm.grow( location=1, kdim=3, split=0.3 )
        code, dims, splim = dtm.encode()
        print( "CODE  ", code, dims, splim )
        print( dtm.fullName( ids=False ) )

        dt2 = DecisionTreeModel( code=code, kdim=dims, split=splim )
        print( "CODE  ", dt2.encode() )
        print( dt2.fullName( ids=False ) )




    def test5( self ):
        print( "  Test 5 DecisionTreeModel" )
        dtm = DecisionTreeModel( ndim=20, kdim=[0],
                    split=0.5, depth=0 )
        dtm.parameters = numpy.arange( dtm.npars, dtype=float ) + 1

        print( dtm.fullName( ids=False ) )
        print( dtm.encode() )
        print( dtm.copy().fullName( ids=False ) )
        print( fmt( dtm.parameters, max=None ) )

        Tools.printclass( dtm )

        print( "GROW    at location 0 dim 1 split 0.34" )
        dtm.grow( offset=0, location=0, kdim=1, split=0.44 )
        print( dtm.fullName() )
        print( dtm.encode() )
        print( dtm.copy().fullName() )
        print( fmt( dtm.parameters, max=None ) )

        print( "GROW    at location 1 dim 1 split 0.44" )
        dtm.grow( offset=0, location=1, kdim=1, split=0.44 )
        print( dtm.fullName( ids=False ) )
        print( dtm.encode() )
        print( dtm.copy().fullName( ids=False ) )
        print( fmt( dtm.parameters, max=None ) )

        print( "GROW    at location 1 dim 2 split 0.54" )
        dtm.grow( offset=0, location=1, kdim=2, split=0.54 )
        print( dtm.fullName() )
        print( dtm.encode() )
        print( dtm.copy().fullName() )
        print( fmt( dtm.parameters, max=None ) )

        print( "GROW    at location 3 dim 3 split 0.64" )
        dtm.grow( offset=0, location=3, kdim=3, split=0.64 )
        print( dtm.fullName() )
        print( dtm.encode() )
        print( dtm.copy().fullName() )
        print( fmt( dtm.parameters, max=None ) )

        dtm.parameters = numpy.arange( dtm.npars, dtype=float ) + 1
        print( fmt( dtm.parameters, max=None ) )

        print( "SHRINK  at location 4" )
        dtm.shrink( offset=0, location=4 )
        print( dtm.fullName( ids=False ) )
        print( dtm.encode() )
        print( dtm.copy().fullName( ids=False ) )
        print( fmt( dtm.parameters, max=None ) )

        print( "SHRINK  at location 0" )
        dtm.shrink( offset=0, location=0 )
        print( dtm.fullName() )
        print( dtm.encode() )
        print( dtm.copy().fullName() )
        print( dtm.copy().encode() )
        print( fmt( dtm.parameters, max=None ) )






    def test6( self ):
        print( "  Test 6 DecisionTreeModel" )
        dtm = DecisionTreeModel( ndim=20, kdim=[0,1,2,3,4,5,6],
                    split=[0.1*k for k in range(2,8)], depth=3 )
        dtm.parameters = numpy.arange( dtm.npars, dtype=float ) + 1

        print( dtm.fullName() )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )


        print( "GROW    at location 2 dim 3 split 0.44" )
        dtm.grow( offset=0, location=2, kdim=3, split=0.44 )
        print( dtm )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )

        print( "VARY    at location 4, dim=5, split=0.55" )
        dtm.vary( location=4, kdim=5, split=0.55 )
        print( dtm )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )

        print( "GROW    at location 8" )
        dtm.grow( offset=0, location=8 )
        print( dtm )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )
        self.assertTrue( len( dtm.parNames ) == dtm.npars )

        print( "SHRINK  at location 7" )
        dtm.shrink( offset=0, location=7 )
        print( dtm )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )

        print( "SHRINK  at location 3" )
        dtm.shrink( offset=0, location=3 )
        print( dtm )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )

        print( "SHRINK  at location 0" )
        dtm.shrink( offset=0, location=0 )
        print( dtm.fullName() )
        print( fmt( dtm.parameters, max=None ) )
        self.assertTrue( dtm.npbase == dtm.npars )
        self.assertTrue( dtm.npars == dtm.countLeaf() )
        self.assertTrue( dtm.ncomp == dtm.countBranch() )

        printclass( dtm )

    def test7( self ):
        print( "  Test 7 DecisionTreeModel" )
        ND = 10
        NP = 100

        dtm = DecisionTreeModel( ndim=ND, kdim=[4,1,1,3,1,4,5],
                    split=0.5, depth=3 )
        print( dtm )

        numpy.random.seed( 12345 )
        xdata = numpy.random.rand( NP, ND )

        ydata = numpy.zeros( NP, dtype=float )
        ydata += 1.4 + 0.3 * xdata[:,4] - 0.1 * xdata[:,1] * xdata[:,6]

        ftr = Fitter( xdata, dtm )

        pars = ftr.fit( ydata )

        print( fmt( pars, max=None ) )

        yfit = dtm( xdata )

        ksrt = numpy.argsort( yfit )
        xxx = numpy.arange( NP, dtype=float )

        if not self.doplot : return

        plt.plot( xxx, ydata[ksrt], 'k.' )
        plt.plot( xxx, yfit[ksrt], 'r-' )

        plt.show()

    def testNestedSampler( self ) :
        print( "========================" )
        print( "  Test Nested Sampler" )
        print( "========================" )

        ND = 10
        NP = 100

        numpy.random.seed( 12345 )
        xdata = numpy.random.rand( NP, ND ) * 10
        print( fmt( xdata ) )

        y = 0.5 * xdata[:,1]
        print( xdata.shape, y.shape )
        y += numpy.where( xdata[:,4] > 3, 5, 0 )
        y += numpy.where( xdata[:,6] > 8, 0, 3 )
        y += numpy.where( xdata[:,4] > 7, 2, 0 )

        noise = numpy.random.randn( NP )
        y += 0.02 * noise

        pm = DecisionTreeModel( ndim=ND, kdim=[0], depth=0 )
        print( pm )

        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ns = NestedSampler( xdata, pm, y, ensemble=100, seed=2031967 )

        ns.verbose = 2
        ns.weed = 10000
        ns.distribution.setLimits( [0.01, 100] )
        if not self.dofull :
            ns.ensemble = 10

        logE = ns.sample( )

        sl = ns.samples

        yfit = sl.average( xdata )

#        ksrt = numpy.argsort( yfit )
        xxx = numpy.arange( NP, dtype=float )

        mi = sl.medianIndex

        print( mi, sl.modusIndex, len( sl ) )

        if not self.doplot : return

        mdl = sl[mi].model
        ksrt = mdl.sortXdata( xdata )
        plt.plot( xxx, y[ksrt], 'k.' )
        plt.plot( xxx, yfit[ksrt], 'r-' )

        for k in range( mi-3, mi+4 ) :
            mdl = sl[k].model
            print( mdl.fullName() )
            ymod = mdl.result( xdata, sl[k].parameters )

            plt.plot( xxx, ymod[ksrt], 'g-' )

        print( ymod )
        print( ksrt )
        print( ymod[ksrt] )

        plt.show()

### TBD

    def XXXtestNestedSampler2( self ) :
        print( "========================" )
        print( "  Test Nested Sampler 2" )
        print( "========================" )

        ND = 10
        NP = 100

        numpy.random.seed( 12345 )
        xdata = numpy.random.rand( NP, ND ) * 10
        print( fmt( xdata ) )

        y = 0.5 * xdata[:,1]
        print( xdata.shape, y.shape )
        y += numpy.where( xdata[:,4] > 3, 5, 0 )
        y += numpy.where( xdata[:,6] > 8, 0, 3 )
        y += numpy.where( xdata[:,4] > 7, 2, 0 )

        noise = numpy.random.randn( NP )
        y += 0.02 * noise

        pm = DecisionTreeModel( ndim=ND, kdim=[0], depth=0 )
        print( pm )

        lolim = [-10]
        hilim = [+10]
        pm.setLimits( lowLimits=lolim, highLimits=hilim )
        Tools.printclass( pm )

        ed = ModelLikelihood( limits=[0.01,100] )

        engs = ["birth", "death", "struct"]
#        engs = ["birth", "struct"]
        ns = NestedSampler( xdata, pm, y, ensemble=100, distribution=ed,
                            engines=engs, seed=2031967 )

#        ns.minimumIterations = 4000
        ns.verbose = 2
#        ns.weed = 10000

        logE = ns.sample( )

        sl = ns.samples

        yfit = sl.average( xdata )

        ksrt = numpy.argsort( yfit )
        xxx = numpy.arange( NP, dtype=float )

        mi = sl.medianIndex

        print( mi, sl.modusIndex, len( sl ) )

        if not self.doplot : return

        plt.plot( xxx, y[ksrt], 'k.' )
        plt.plot( xxx, yfit[ksrt], 'r-' )

        for k in range( mi-3, mi+4 ) :
            mdl = sl[k].model
            print( mdl.fullName() )
            ymod = mdl.result( xdata, sl[k].parameters )

            plt.plot( xxx, ymod[ksrt], 'g-' )

        print( ymod )
        print( ksrt )

        plt.show()

        print( ymod[ksrt] )

    def suite( cls ):
        return unittest.TestCase.suite( TestDynamicModel.__class__ )


if __name__ == '__main__':
    unittest.main()




#run with: python3 -m unittest TestDimensions

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

#from BayesicFitting import PolynomialModel, SineModel
#from Model import Model

from StdTests import stdModeltest as stdMtest

from BayesicFitting import formatter as fmt
from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ) :
        model = PolynomialModel( 2 )
        print( "====Test1===== dimensions  ", model.ndim, model.ndout )

        print( model )

        p = numpy.asarray( [0.0, 0.4, 0.2] )
        x = numpy.linspace( 0, 10, 11, dtype=float )

        stdMtest( model, p, x=x, plot=self.doplot )
        self.stdModeltest( model, p, x=x, plot=self.doplot )
        self.stdModeltest( model, p, x=x[[0]], plot=self.doplot )

    def test2( self ) :
        model = PolySurfaceModel( 2 )
        print( "====Test2===== dimensions  ", model.ndim, model.ndout )

        print( model )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0] )
        x = numpy.linspace( 0, 10, 22, dtype=float ).reshape( (11,2) )

        self.stdModeltest( model, p, x=x, plot=self.doplot )
        self.stdModeltest( model, p, x=x[[0],:], plot=self.doplot )

    def test3( self ) :
        model = StellarOrbitModel( )
        print( "====Test3===== dimensions  ", model.ndim, model.ndout )

        print( model )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0, 2.0] )
        x = numpy.linspace( 0, 10, 11, dtype=float )

        self.stdModeltest( model, p, x=x, plot=self.doplot )
        self.stdModeltest( model, p, x=x[[0]], plot=self.doplot )

    def test4( self ) :
        model = SoftMaxModel( ndim=3, ndout=2 )
        print( "====Test4===== dimensions  ", model.ndim, model.ndout )

        print( model )

        p = numpy.asarray( [0.0, 0.4, 0.2, 1.0, 0.0, 1.0] )
        x = numpy.linspace( 0, 10, 33, dtype=float ).reshape( (11,3) )

        self.stdModeltest( model, p, x=x, plot=self.doplot )
        self.stdModeltest( model, p, x=x[[0],:], plot=self.doplot )

    def test5( self ) :
        model = Kernel2dModel( )
        print( "====Test5===== dimensions  ", model.ndim, model.ndout )

        print( model )

        p = numpy.asarray( [0.4, 0.2, 1.0, 2.0] )
        x = numpy.linspace( 0, 10, 22, dtype=float ).reshape( (11,2) )

        self.stdModeltest( model, p, x=x, plot=self.doplot )

    def stdModeltest( self, model, p, x, plot=False ) :

        print( "result:" )
        print( fmt( model.result( x, p ) ) )
        print( "derivative:" )
        print( fmt( model.derivative( x, p ) ) )
        print( "strict:" )
        print( fmt( model.strictNumericDerivative( x, p ) ) )
        print( "partial:" )
        print( fmt( model.partial( x, p ) ) )
        print( "strict:" )
        print( fmt( model.strictNumericPartial( x, p ) ) )
#        print( "numeric:" )
#        print( fmt( model.numPartial( x, p ) ) )



    def suite( cls ):
        return unittest.TestCase.suite( TestDimensions.__class__ )


if __name__ == '__main__':
    unittest.main()




from .Model import Model
from . import Tools
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class NonLinearModel( Model ):
    """
    Anchestor of all non-linear models.

    The general non-linear model needs both the function value and the partials.

    It contains provisions for mixed models. (TBC)

    Attributes
    ----------
    _linear : list of int
         list of indices for the linear parameters (in case of a mixed model)

    Attributes from Model
    ---------------------
        parameters, stdevs, npchain
        _next, _head, _operation
        xUnit, yUnit (relegated to model)

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Author       Do Kester

    """
    def __init__( self, nparams, ndim=1, copy=None, **kwargs ):
        """
        Parent class for all non linear models.

        Parameters
        ----------
        nparams : int
            number of parameters in the model
        ndim : int
            dimensionality of the input. Default 1
        copy : NonLinearModel
            model to be copied.
        kwargs : dict
            Possibly includes keywords from
                @FixedModel :   fixed, names
                @BaseModel :    posIndex, nonZero

        """
        super( NonLinearModel, self ).__init__( nparams, ndim=ndim, copy=copy, **kwargs )
        if copy is None :
            setatt( self, "_linear", set() )
        else :
            setatt( self, "_linear", copy._linear.copy() )

    def setMixedModel( self, lindex ):
        """
        Convert a NonLinear model into a Mixed model with linear and
        non-linear parameters.

        Reset with SetMixedModel( null );

        Parameters
        ----------
        lindex : list of int
            indices of the linear parameters

        """
        if lindex is None:
            setatt( self, "_linear", set() )
        else:
            setatt( self, "_linear", set( lindex ) )

    def isMixed( self ):
        """ Returns true when linear indices have been set  """
        return len( self._linear ) > 0

    def getNonLinearIndex( self ):
        """ Returns the index of the non-linear parameters.  """

        np = self.getNumberOfParameters()
        return set( [x for x in range( np )] ) - self._linear

    def partial( self, xdata, param=None, useNum=False ):
        """
        Return the partial derivatives for the model.

        Parameters
        ----------
        xdata : array_like
            the value at which to calculate the partials
        param : array_like
            the parameters of the model. Default the self.parameters
        useNum : boolean
            if True use numeric partial derivatives. Default False

        """
        if param is None :
            param = self.parameters
        return super().partial( xdata, param, useNum  )




import numpy as numpy
import math
import sys

#from .ErrorDistribution import ErrorDistribution
from .ScaledErrorDistribution import ScaledErrorDistribution
from .LinearModel import LinearModel
from .Fitter import Fitter
from .LevenbergMarquardtFitter import LevenbergMarquardtFitter
#from .NestedSampler import NestedSampler
from .Formatter import formatter as fmt
from .Tools import setAttribute as setatt


__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2021 Do Kester


class ModelDistribution( ScaledErrorDistribution ):
    """
    To calculate the probability of a model M from a set of models S,
    given some data D, use Bayes rule:

        P( M|DS ) = P( M|S ) * P( D|MS ) / P( D|S )
        posterior = prior   * likelihood / evidence

    This class calculates the likelihood P( D|MS ).
    On another level where we calculate the probability of the
    parameters p, we see this likelhood appear as evidence P( D|M ).

    Again using Bayes :

        P( p|DM ) = P( p|M ) * P( D|pM ) / P( D|M )

    The evidence here is calculated as the integral over a Gausian
    approximation of the posterior.

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, arbiter=None, scale=1.0, limits=None, copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        arbiter : None or BaseFitter or str
            to provide the evidence
            None    select fitter automatically
            BaseFiter   Use this fitter
            str     "fitter", "levenberg", "curve", "amoeba"

        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.

        copy : ModelDistribution
            distribution to be copied.

        """
        setatt( self, "arbiter", arbiter )

#        setatt( self, "limits", limits )
#        setatt( self, "scale", scale )

        super( ).__init__( limits=limits, scale=scale, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return ModelDistribution( copy=self, arbiter=self.arbiter )
#        return ModelDistribution( copy=self, arbiter=self.arbiter, scale=self.scale, limits=self.limits )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True


    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
            Return optimal parameters of the fit

        """
        mdl = problem.model

        try :
            noiselim = self.hyperpar[0].getLimits()
            fscale = None
        except :
            noiselim = None
            fscale = self.scale

        if noiselim is None and mdl.npars < len( allpars ) :
            allpars[-1] = self.hyperpar[0].hypar

#        print( "MD   ", noiselim, fmt( allpars, max=None ) )


        if self.arbiter is None :
            if isinstance( mdl, LinearModel ) :
                ftr = Fitter( problem.xdata, mdl, fixedScale=fscale )
            else :
                ftr = LevenbergMarquardtFitter( problem.xdata, mdl, fixedScale=fscale )
        elif isinstance( self.abriter, BaseFitter ) :
            ftr = self.arbiter

        elif not isinstance( self.arbiter, str ) :
            raise ValueError( "Cannot interpret %s as string or fitter" % self.arbiter )

        else :
            name = str.lower( self.arbiter )

            if name == "fitter" :
                ftr = Fitter( problem.xdata, mdl, fixedScale=fscale )
            elif name.startswith( "leven" ) :
                ftr = LevenbergMarquardtFitter( problem.xdata, model, fixedScale=fscale )
            elif name.startswith( "curve" ) :
                ftr = CurveFitter( problem.xdata, model, fixedScale=fscale )
            elif name.startswith( "amoeba" ) :
                ftr = AmoebaFitter( problem.xdata, model, fixedScale=fscale )

#               produces circular imports
#            elif name.startswith( "nested" ) :
#                ns = NestedSampler( problem=problem, limits=noiselim, verbose=0 )
#                evidence = ns.sample()
#                allpars[:mdl.npars] = pars
#                if noiselim is not None :
#                    allpars[-1] = ftr.scale
#                return evidence

            else :
                raise ValueError( "Cannot interpret %s" % self.arbiter )

#        print( "MD Fitr    ", ftr ) 

        try :
            pars = ftr.fit( problem.ydata )
            evidence = ftr.getLogZ( limits=mdl.getLimits(), noiseLimits=noiselim )

            allpars[:mdl.npars] = pars
            if noiselim is not None :
                allpars[-1] = ftr.scale

#            print( problem.npars, mdl.npars, noiselim, ftr.scale, allpars )

        except numpy.linalg.LinAlgError :
            evidence = - sys.float_info.max
        except ValueError :                         # math domain error
            evidence = - sys.float_info.max


        return evidence

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual
           
        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem   
        mockdata : array_like
            as calculated by the model
        
        """
        return self.logLikelihood_alt( problem, allpars )


    def __str__( self ) :
        return "ModelDistribution"



import numpy as numpy
from astropy import units
import math
from . import Tools

from .OrderEngine import OrderEngine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2022 Do Kester

class StartOrderEngine( OrderEngine ):
    """
    StartEngine generates a parameter list in random order.

    It is used to initialize the set of trial samples.

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.
        Parameters
        ----------
        copy : StartEngine
            engine to be copied

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return StartOrderEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "StartOrderEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, fitIndex=None ):
        """
        Execute the engine by a random selection of the parameters.

        Parameters
        ----------
        kw : id
            id of waker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        fitIndex : array_like
            list of parameter indices (not active)
        Returns
        -------
        int : the number of successfull moves

        """
        walker = self.walkers[kw]

        problem = walker.problem
#        print( walker.allpars )
        par = numpy.random.permutation( walker.allpars )
#        print( par )

        logL = self.errdis.logLikelihood( problem, par )
        self.setWalker( walker.id, problem, par, logL )

        return len( par )





# run with : python3 -m unittest TestWeights

import unittest
import numpy as np
import math
import os

from numpy.testing import assert_array_equal as assertAE
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt
from BayesicFitting import *
from BayesicFitting import formatter as fmt


class Test( unittest.TestCase  ) :

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testWeights1( self ) :
        print( "====testWeights 1====================" )

        self.stdtst1( Fitter )

    def testWeights2( self ) :
        print( "====testWeights 1====================" )

        self.stdtst2( Fitter, verbose=0 )

    def testWeights3( self ) :
        print( "====testWeights 3====================" )

        self.stdtst1( LevenbergMarquardtFitter )

    def testWeights4( self ) :
        print( "====testWeights 4====================" )

        self.stdtst1( QRFitter )

    def testWeights5( self ) :
        print( "====testWeights 5====================" )

        self.stdtst1( CurveFitter )

### Test is OK but fails exact outcomes.
### because we need order = 1 on AmoebaFitter.
#    def testWeights6( self ) :
#        print( "====testWeights 6====================" )
#
#        self.stdtst1( AmoebaFitter, order=1 )


    def testPlot( self ) :
        if not self.doplot :
            return

        nn = 5
        x = np.arange( nn, dtype=float )
        y = 2 * ( x % 2 ) - 1
        y[2] = 0.0
        x4 = np.arange( 4 * nn, dtype=float )
        y4 = np.append( y, y )
        y4 = np.append( y4, y4 )

        upr = UniformPrior( limits=[-10,10] )

        pm = PolynomialModel( 0 )
        pm.priors = upr

        bf = Fitter( x, pm )
        par = bf.fit( y )
        evi = bf.evidence

        print( "evidence   %8.3f" % evi )

        s = 0
        ns = 10
        sam = numpy.zeros( ns, dtype=float )
        for k in range( ns ):
            evid1, prec1, logL1 = self.nstst( x, y, pm, seed=2345+k*56 )
            print( "%4d  NSevid  %8.3f +- %.3f" % ( k, evid1, prec1 ) )
            sam[k] = evid1
            s += evid1
        aver = s / ns
        print( "aver   %8.3f" % aver )

        nbin = 5
        plt.hist( sam, nbin, facecolor='g', alpha=0.5 )
        plt.plot( [evi,evi], [0, ns/nbin], 'r-' )
        plt.plot( [aver,aver], [0, ns/nbin], 'b-' )
        plt.show()

    def nstdtst( self, x, y, mdl, ftr, wgt=None, acc=None, verbose=0 ) :
        par = ftr.fit( y, weights=wgt, accuracy=acc )
        std = ftr.stdevs
        cov = ftr.covariance
        hes = ftr.hessian
        var = ftr.makeVariance()
        chi = ftr.chisq
        scl = ftr.scale
        swt = ftr.sumwgt

        print( "pars  ", par, "stdv  ", std, "hess  ", hes, "cov  ", cov )
        print( "chisq  %8.3f  scale   %8.3f sumwgt  %8.3f" % ( chi, scl, swt ) )

        assertAE( par, ftr.parameters )
        assertAE( std, np.sqrt( np.diag( cov ) ) )
        assertAE( scl, np.sqrt( var ) )

        evi = ftr.evidence
        lgl = ftr.logLikelihood
        lgo = ftr.logOccam
        print( "Evid    %8.3f  logL    %8.3f logOcc    %8.3f" % ( evi, lgl, lgo ) )


        ns = NestedSampler( x, mdl, y, weights=wgt, accuracy=acc, verbose=verbose, 
                            limits=[0.01,10] )

        nevi = ns.sample()

        npar = ns.parameters
        nstd = ns.stdevs
        nlgl = ns.walkers[0].logL
        prec = ns.precision 

        print( "NSpars  ", ns.parameters, "stdv  ", ns.stdevs )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( nevi, prec, lgl ) )

        assertAAE( par, npar )
        assertAAE( std, nstd )
        assertAAe( lgl, nlgl )
        self.assertTrue( abs( evi - nevi ) < 2 * prec )

        return ( par, std, scl, var, chi, swt )



    def stdtst( self, y, pm, bf, wgt=None, acc=None ) :
        par = bf.fit( y, weights=wgt, accuracy=acc )
        std = bf.stdevs
        cov = bf.covariance
        hes = bf.hessian
        var = bf.makeVariance()
        chi = bf.chisq
        scl = bf.scale
        swt = bf.sumwgt

        print( "pars  ", par, "stdv  ", std, "hess  ", hes, "cov  ", cov )
        print( "chisq  %8.3f  scale   %8.3f sumwgt  %8.3f" % ( chi, scl, swt ) )

        assertAE( par, bf.parameters )
        assertAE( std, np.sqrt( np.diag( cov ) ) )
        assertAE( scl, np.sqrt( var ) )

        if pm.hasPriors() :
            evi = bf.evidence
            print( "Evid    %8.3f  logL    %8.3f logOcc    %8.3f" % ( evi, bf.logLikelihood, bf.logOccam ) )

        return ( par, std, scl, var, chi, swt )

    def nstst( self, x, y, mdl, wgt=None, acc=None, verbose=0, seed=2023 ) :
        ns = NestedSampler( x, mdl, y, weights=wgt, accuracy=acc, verbose=verbose, seed=seed )
        evi = ns.sample()

        print( "NSpars  ", ns.parameters, "stdv  ", ns.stdevs )
        
        return ( evi, ns.precision, ns.walkers[0].logL )


    def stdtst2( self, myFitter, verbose=0 ) :
        nn = 5
        x = np.arange( nn, dtype=float )
        y = 2 * ( x % 2 ) - 1
        y[2] = 0.0
        x4 = np.arange( 4 * nn, dtype=float )
        y4 = np.append( y, y )
        y4 = np.append( y4, y4 )

        upr = UniformPrior( limits=[-10,10] )

        pm = PolynomialModel( 0 )
        pm.priors = upr
        bf = myFitter( x, pm )

        print( "\n    %s" % bf )
    
        wgt = None
        print( "\n======= 5 data points; weighs = ", wgt )
        print( "ydata  ", y )
        ( par1, std1, scl1, var1, chi1, swt1 ) = self.stdtst( y, pm, bf, wgt=wgt )
        evi1 = bf.evidence
        evid1, prec1, logL1 = self.nstst( x, y, pm, wgt=wgt, verbose=verbose )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( evid1, prec1, logL1 ) )

        print( "\n======= 20 data points; weights = ", wgt )
        print( "ydata  ", y4 )
        pm = PolynomialModel( 0 )
        pm.priors = upr
        bf = myFitter( x4, pm )
        ( par2, std2, scl2, var2, chi2, swt2 ) = self.stdtst( y4, pm, bf )
        evid2, prec2, logL2 = self.nstst( x4, y4, pm, verbose=verbose )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( evid2, prec2, logL2 ) )

        sq15 = 1.0 / math.sqrt( 1.5 )
        acc = np.asarray( [2, 0.5, sq15, 2.0, 0.5], dtype=float )
        wgt = 1.0 / ( acc * acc )
        # normalize to weights to nn (=5)
        # making case 3 completely the same as case 1 (also for the evidence
        wgt *= 5 / np.sum( wgt )
        acc = 1 / np.sqrt( wgt )
        y *= acc


        print( "\n======= 5 data points; weights = ", fmt( wgt, max=None ) )
        print( "ydata  ", y )
        pm = PolynomialModel( 0 )
        pm.priors = upr
        bf = myFitter( x, pm )
        ( par3, std3, scl3, var3, chi3, swt3 ) = self.stdtst( y, pm, bf, wgt=wgt )
        evi3 = bf.evidence
        evid3, prec3, logL3 = self.nstst( x, y, pm, wgt=wgt, verbose=verbose )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( evid3, prec3, logL3 ) )

        print( "\n======= 5 data points; accuracy = ", fmt( acc, max=None ) )
        print( "ydata  ", y )
        pm = PolynomialModel( 0 )
        pm.priors = upr
        bf = myFitter( x, pm )
        ( par4, std4, scl4, var4, chi4, swt4 ) = self.stdtst( y, pm, bf, acc=acc )
        evid4, prec4, logL4 = self.nstst( x, y, pm, acc=acc, verbose=verbose )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( evid4, prec4, logL4 ) )

        ## all parameters are the same
        assertAAE( par1, par2 )
        assertAAE( par3, par4 )
        assertAAE( par1, par4 )

        assertAAE( evi1, evi3 )

        assertAAE( chi1, chi3 )
        assertAAE( chi1, chi4 )
        assertAAE( scl1, scl3 )
        assertAAE( scl1, scl4 )
        assertAAE( swt1, swt3 )
        assertAAE( swt1, swt4 )
        assertAAE( std1, std3 )
        assertAAE( std1, std4 )


    def stdtst1( self, myFitter, order=0 ) :
        nn = 5
        x = np.arange( nn, dtype=float )
        y = x % 2
        y[2] = 0.5
        x4 = np.arange( 4*nn, dtype=float )
        y4 = x4 % 2
        y4[range(2,20,5)] = 0.5

        upr = UniformPrior( limits=[-10,10] )
#        printclass( upr )

        pm = PolynomialModel( order )
        pm.priors = upr
#        printclass( pm )


        bf = myFitter( x, pm )

        print( "\n    %s" % bf )
    
        print( "\n======= 5 data points; no weights ===" )
        ( par1, std1, scl1, var1, chi1, swt1 ) = self.stdtst( y, pm, bf )
        evid1, prec1, logL1 = self.nstst( x, y, pm )
        print( "NSevid  %8.3f +- %.3f   logL %.3f" % ( evid1, prec1, logL1 ) )

#        ( par1, std1, scl1, var1, chi1, swt1 ) = self.nstdtst( x, y, pm, bf )

        print( "\n======= 20 data points; no weights ===" )
        pm = PolynomialModel( order )
        pm.priors = upr
        bf = myFitter( x4, pm )
        ( par2, std2, scl2, var2, chi2, swt2 ) = self.stdtst( y4, pm, bf )

        print( "\n======= 5 data points; weights = 4 ===" )
        pm = PolynomialModel( order )
        pm.priors = upr
        bf = myFitter( x, pm )
        w = np.zeros( nn, dtype=float ) + 4
        ( par3, std3, scl3, var3, chi3, swt3 ) = self.stdtst( y, pm, bf, wgt=w )

        print( "\n======= 5 data points; accuracy = 0.5 ===" )
        pm = PolynomialModel( order )
        pm.priors = upr
        bf = myFitter( x, pm )
        acc = np.ones( nn, dtype=float ) / 2
#        acc = np.ones( nn, dtype=float )
        print( "ydata  ", y )
        ( par4, std4, scl4, var4, chi4, swt4 ) = self.stdtst( y, pm, bf, acc=acc )
        evid4, prec4, logL4 = self.nstst( x, y, pm, acc=acc )
        print( "NSevid  %8.3f +- %.3f    logL %.3f" % ( evid4, prec4, logL4 ) )

        ## all parameters are the same
        assertAAE( par1, par2 )
        assertAAE( par3, par4 )
        assertAAE( par1, par4 )

        ## 2 and 3 are completely equal
        assertAAE( std2, std3 )
        assertAAE( chi2, chi3 )
        assertAAE( scl2, scl3 )
        assertAAE( var2, var3 )
        assertAAE( swt2, swt3 )

        assertAAE( std1, std4 )
        assertAAE( scl1 * 2, scl4 ) 




        





if __name__ == '__main__':
    unittest.main( )




import math

from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.6.2"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class UniformPrior( Prior ):
    """
    Uniform prior distribution, for location parameters.

    A uniform prior is a improper prior ( i.e. its integral is unbound ).
    Because of that it always needs limits, low and high, such that
    -Inf < low < high < +Inf.

        Pr( x ) = 1 / ( high - low )    if low < x < high
                  0                     elsewhere

    domain2Unit: u = ( d - lo ) / range
    unit2Domain: d = u * range + lo

    Examples
    --------
    >>> pr = UniformPrior()                                 # unbound prior
    >>> pr = UniformPrior( limits=[0,10] )                  # limited to the range [0,10]
    >>> pr = UniformPrior( circular=math.pi )               # circular between 0 and pi
    >>> pr = UniformPrior( limits=[2,4], circular=True )    # circular between 2 and 4

    Attributes
    ----------
    _range : float
        highlimit - lowlimit

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain


    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, circular=False, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        limits : None or [float,float]
            None    no limits are set
            2 floats    lowlimit and highlimit
        circular : bool or float
            True : circular with period from limits[0] to limits[1]
            float : period of circularity
        prior : UniformPrior
            to be copied
        """
        super( ).__init__( limits=limits, circular=circular, prior=prior )

    def copy( self ):
        """ Return a (deep) copy of itself. """
        return UniformPrior( prior=self, limits=self.limits, circular=self.circular )

    def getIntegral( self ) :
        """
        Return integral of UniformPrior from lowLimit to highLimit.
        """
        return self._urng

    def domain2Unit( self, dval ):
        """
        Return the dval as uval

        In Prior.limitedDomain2Unit the dval is transformed into a uval

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return dval

    def unit2Domain( self, uval ):
        """
        Return the uval as dval

        In Prior.limitedUnit2Domain the uval is transformed into a dval

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        return uval

    def result( self, x ):
        """
        Return a the result of the distribution function at x.

        Parameters
        ----------
        x : float
            value within the domain of a parameter

        """
        if math.isinf( self._urng ) :
            raise AttributeError( "Limits are needed for UniformPrior" )

        return 0.0 if self.isOutOfLimits( x ) else 1.0 / self._urng


# logResult has no better definition than the default: just take the math.log of result.
# No specialized method here.

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.

        Parameters
        ----------
        p : float
            the value

        """
        return math.nan if self.isOutOfLimits( p ) else 0

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return self.hasLowLimit( ) and self.hasHighLimit( )

    def shortName( self ):
        """ Return a string representation of the prior.  """
        return str( "UniformPrior" + ( " unbound." if not self.isBound( ) else "" ) )







import numpy as numpy
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2022 Do Kester

class PolySurfaceModel( LinearModel ):
    """
    General polynomial surface model of arbitrary degree.

        f( x,y:p ) = &sum;_d &sum;_k p_n * x^{d-k} * y^k )

    where the first sum is over d running from 0 to degree ( inclusive )
    and the second sum is over k running from 0 to d ( inclusive ).
    The index n is just incrementing, making all p's different.

    It is a 2-dimensional linear model.

    Examples
    --------
    poly = PolySurfaceModel( 3 )         # 3rd degree polynomial
    print poly.getNumberOfParameters( )        # 10

    Author      Do Kester

    Attributes
    ----------
    degree : int
        degree of the polynomial

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Polynominal surface of a certain degree. Two dimensions.

        degree      polysurface
          0         p_0
          1         p_0 + p_1 * x + p_2 * y
          2         p_0 + p_1 * x + p_2 * y + p_3 * x^2 + p_4 * x * y + p_5 * y^2
          3         p_0 + p_1 * x + p_2 * y + p_3 * x^2 + p_4 * x * y + p_5 * y^2 +
                        p_6 * x^3 + p_7 * x^2 * y + p_8 * x * y^2 + p_9 * y^3
        etc.

        The number of parameters is ( degree+2 ) * ( degree+1 ) / 2

        Parameters
        ----------
        degree : int
            the degree of the polynomial.
        copy : PolySurfaceModel
            model to be copied
        fixed : dictionary of {int:float|Model}
            int             list if parameters to fix permanently. Default None.
            float|Model     list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See @FixedModel

        """
        super( PolySurfaceModel, self ).__init__( (degree+2)*(degree+1)//2, ndim=2,
                copy=copy, **kwargs )
        self.degree = degree

        if copy is not None and hasattr( copy, "xindex" ) :
            setatt( self, "xindex", copy.xindex )
            setatt( self, "yindex", copy.yindex )
            setatt( self, "xfact", copy.xfact )
            setatt( self, "yfact", copy.yfact )
            

    def copy( self ):
        """ Copy method.  """
        return PolySurfaceModel( self.degree, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: degree
        """
        if name == 'degree' :
            setatt( self, name, value, type=int )
        else :
            super( PolySurfaceModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input values.

        The partials are the powers of x,y ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        nx = Tools.length( xdata[:,0] )
        part = numpy.zeros( ( nx, self.npmax ), dtype=float )
        n = 0
        for d in range( self.degree + 1 ) :
            x = numpy.ones( nx, dtype=float )
            i = 0
            while i <= d :
                part[:,n+i] = x
                x *= xdata[:,1]
                i += 1
            x = numpy.ones( nx, dtype=float )
            i = d
            while i >= 0 :
                part[:,n+i] *= x
                x *= xdata[:,0]
                i -= 1
            n += d + 1
        if parlist is None :
            return part

        return part[:,parlist]

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each input (=x).

        degree      df/dx
          0         0
          1         p_1
          2         p_1 + 2 * p_3 * x + p_4 * y
          3         p_1 + 2 * p_3 * x + p_4 * y + 3 * p_6 * x^2 + 2 * p_7 * x * y + p_8 * y^2

        degree      df/dy
          0         0
          1         p_2
          2         p_2 + p_4 * x + 2 * p_5 * y
          3         p_2 + p_4 * x + 2 * p_5 * y + p_7 * x^2 + 2 * p_8 * x * y + 3 * p_9 * y^2


        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            parameters for the model.

        """
        if self.npmax <= 1 :
            return numpy.zeros_like( xdata )

        np = self.npmax

        if not hasattr( self, "xindex" ) :
            indx = []
            fact = []
            n = 1
            k = 1
            while k < np :
#                print( np, k, n )
                indx += [i for i in range( k, k+n )]
                fact += [n-f for f in range( n )]
                k += n + 1
                n += 1

#                print( "X  ", fmt( indx, max=None ), fmt( fact, max=None ) )

            setatt( self, "xindex", indx )
            setatt( self, "xfact", fact )

            indx = []
            fact = []
            k = 2
            n = 1
            while k < np :
#                print( np, k, n )
                indx += [i for i in range( k, k+n )]
                fact += [f+1 for f in range( n )]
                k += n + 1
                n += 1

#                print( "Y  ", fmt( indx, max=None ), fmt( fact, max=None ) )

            setatt( self, "yindex", indx )
            setatt( self, "yfact", fact )


        p = params[self.xindex] * self.xfact
        deg  = self.degree - 1

        dfdx = PolySurfaceModel( deg ).result( xdata, p )

        p = params[self.yindex] * self.yfact
        dfdy = PolySurfaceModel( deg ).result( xdata, p )

        return numpy.append( dfdx, dfdy ).reshape( 2, -1 ).T


    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "PolySurface: f( x,y:p ) ="
        n = 0
        plus = ""
        for d in range( self.degree + 1 ) :
            i = d
            k = 0
            while k <= d :
                bn += "%s p_%d"%(plus,n)
                if i > 0 :
                    bn += " * x%s"%( "" if i == 1 else "^%d"%i )
                if k > 0 :
                    bn += " * y%s"%( "" if k == 1 else "^%d"%k )
                n += 1
                k += 1
                i -= 1
            plus = " +"
        return bn

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters  k    parameter number.

        """
        dk = 0
        while k > dk:
            dk += 1
            k -= dk
        return self.yUnit / ( ( self.xUnit[0]**k ) * ( self.xUnit[1] ** (dk-k) ) )





## run as : python3 -m unittest TestErrorDistribution

from __future__ import print_function

import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
import unittest
import os
from astropy import units
import math
import sys
import matplotlib.pyplot as plt
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt
#from MultiDimAverageModel import MultiDimAverageModel
#from BoxedErrorDistribution import BoxedErrorDistribution

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2017 - 2021  Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2 )
    EPSILON = 1e-6

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define y dependent variable
    data = numpy.asarray( [-11.1, -9.2, -5.1, -5.2, -1.1, 1.2, 1.1, 5.2, 5.1, 8.2, 11.2], dtype=float )

    # Define wgt weigth
    wgt = numpy.asarray( [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=float )

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        numpy.set_printoptions(formatter={'float': '{: 0.3f}'.format})


    def testGaussErrorDistribution( self ):

        print( "=======   Test Gauss Error Distribution  ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10,1], dtype=float )
        ged = GaussErrorDistribution( )
        self.assertTrue( ged.acceptWeight() )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data,
                weights=self.wgt )

        fitIndex = [0,1,-1]

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        chisq = ged.getChisq( problem, param )
        print( "chisq = %8.3f"%( chisq ) )
        lLdata = ged.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )

        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f  %8.3f" % ( logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        scale = 0.1
        param[2] = scale
        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
#        fitIndex = numpy.arange( 3 )
        fitIndex = numpy.asarray( [0,1,-1] )
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        print( "Using scale = 0.5 and weights." )
        scale = 0.5
        param[2] = scale
        problem.weights = self.wgt

        logL = ged.logLikelihood( problem, param )
        print( "logL = %8.3f  %8.3f" % ( logL, logL0 ) )

        lLdata = ged.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        assertAAE( logL, logL0 )

        dL = ged.partialLogL( problem, param, fitIndex )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 1.0
        for i in range( 11 ) :
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ged.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def constrainBox( self, logL, problem, allpars, logLlow ):
        box = [-2,2]
        for p in allpars[:-1] :
            if ( p < box[0] ) or ( p > box[1] ) :
                return logLlow - 1

        return logL

    def XXXtestGaussErrorDistribution1( self ):

        print( "=======   Test Gauss Error Distribution 1 ==================" )

        ndim = 2
        poly = MultiDimAverageModel( ndim )
        param = numpy.asarray( [0,0,1], dtype=float )
        ged = GaussErrorDistribution( )
        ged.constrain = self.constrainBox
        self.assertTrue( ged.acceptWeight() )

        numpy.random.seed( 2345 )
        ydata = numpy.random.randn( ndim * 11 ).reshape( 11, ndim )

        print( fmt( ydata ) )
        problem = MultipleOutputProblem( model=poly, xdata=self.x, ydata=ydata )
#                weights=self.wgt )

        fitIndex = [0,1,-1]

        print( fmt( poly.result( self.x, param[:ndim] ) ) )


        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        chisq = ged.getChisq( problem, param )
        print( "chisq = %8.3f"%( chisq ) )

        pp = numpy.array( [0,0,1], dtype=float )
        print( " 0 0: ", fmt( ged.logLdata( problem, pp ), max=None ) )
        pp = numpy.array( [0,1,1], dtype=float )
        print( " 0 1: ", fmt( ged.logLdata( problem, pp ), max=None ) )
        pp = numpy.array( [1,1,1], dtype=float )
        print( " 1 1: ", fmt( ged.logLdata( problem, pp ), max=None ) )
        pp = numpy.array( [1,0,1], dtype=float )
        print( " 1 0: ", fmt( ged.logLdata( problem, pp ), max=None ) )



        lLdata = ged.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )

        ged.lowLhood = -math.inf        ## is normally set by NS
        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f  %8.3f" % ( logL, logL0, altL ) )
#        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        scale = 0.1
        param[2] = scale
        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
#        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
#        fitIndex = numpy.arange( 3 )
        fitIndex = numpy.asarray( [0,1,-1] )
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
#        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        print( "Using scale = 0.5 and weights." )
        scale = 0.5
        param[2] = scale

        wgts = []
        for k in range( ndim ) :
            wgts = numpy.append( wgts, self.wgt )

        problem.weights = wgts

        logL = ged.logLikelihood( problem, param )
        print( "logL = %8.3f  %8.3f" % ( logL, logL0 ) )

        lLdata = ged.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        assertAAE( logL, logL0 )

        dL = ged.partialLogL( problem, param, fitIndex )
#        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        print( ged.logLikelihood )
        scale = 1.0
        for i in range( 11 ) :
            param = numpy.asarray( [i-5,-4,1], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ged.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def testGaussErrorDistribution2( self ):
        print( "\n=====   Test 2 Gauss Error Distribution====================" )
        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )
        param = numpy.asarray( [5, 1, -0.3, 1, 1, 1], dtype=float )
        data2 = self.data.copy( ) * 0.1
        data2[1:5] = [2.3, 5.4, 4.5, 2.0]
        print( data2 )

        problem = ClassicProblem( model=gm, xdata=self.x, ydata=self.data )
        fitIndex = [0,1,2,3,4,-1]

        ged = GaussErrorDistribution( )
        Tools.printclass( ged )
        np = gm.npchain
        chisq = ged.getChisq( problem, param )
        print( "chisq = %8.3f"%( chisq ) )
        logL = ged.logLikelihood( problem, param )
        altL = ged.logLikelihood_alt( problem, param )

        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 0.1
        param[np] = scale
        logL = ged.logLikelihood( problem, param )
        altL = ( -11 * ( 0.5 * math.log( 2 * math.pi ) +
                    math.log( scale ) ) - 0.5 * chisq / ( scale * scale ) )
        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[np] = scale
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        fitIndex = numpy.asarray( [0,2,3,4] )
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,3,4,-1] )
        problem.weights = self.wgt
        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem,  param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,4,-1] )

        ged.hyperpar[0] = NoiseScale()
        Tools.printclass( ged )
        ged.keepFixed( fixed={0:2.5} )
        Tools.printclass( ged )
        Tools.printclass( ged.hyperpar[0] )

        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

    def testExponentialErrorDistribution1( self ):

        print( "=======   Test Exponential Error Distribution 1 ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10], dtype=float )
        ged = LaplaceErrorDistribution( )
        ggd = ExponentialErrorDistribution( limits=[[0.1,0.5],[10,3]] )
        self.assertTrue( ggd.acceptWeight() )

        self.assertTrue( ggd.hyperpar[0].prior.lowLimit == 0.1 )
        self.assertTrue( ggd.hyperpar[0].prior.highLimit == 10 )
        self.assertTrue( ggd.hyperpar[1].prior.lowLimit == 0.5 )
        self.assertTrue( ggd.hyperpar[1].prior.highLimit == 3 )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )
        fitIndex = [0,1,2,3,4,-1]


        power = 1.0
        ggscl = 1.0
        lpscl = 1.0
        hypar = [ggscl, power]
        ggsigma = ggd.toSigma( hypar )
        lpsigma = ged.toSigma( lpscl )
        sigma = ggsigma / lpsigma
        print( ggsigma, lpsigma, sigma )

        ff = 1.0 / sigma
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [lpscl] )
        res = problem.residuals( param )
        chisq = ged.getSumRes( problem, lppar )
        chipp = ggd.getChipow( problem, ggpar )
        print( "chipp = %8.3f   chisq = %8.3f"%( chipp, chisq ) )
        assertAAE( chisq, chipp )

        ggfi = numpy.asarray( [0,1,-2,-1] )
        lpfi = numpy.asarray( [0,1,-1] )
        logL = ged.logLikelihood( problem, lppar )
        lggL = ggd.logLikelihood( problem, ggpar )
        altL = ggd.logLikelihood_alt( problem, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f   alt  = %8.3f" % ( lggL, logL, altL ) )
        assertAAE( logL, lggL )
        assertAAE( altL, lggL )

        lLdata = ggd.logLdata( problem, ggpar )
        logL0 = numpy.sum( lLdata )
        print( "logL  = %8.3f   ldata = %8.3f" % (logL, logL0 ) )
        assertAAE( logL, logL0 )


        dL = ged.partialLogL( problem, lppar, lpfi )
        nL = ged.numPartialLogL( problem, lppar, lpfi )
        dG = ggd.partialLogL( problem, ggpar, ggfi)
        nG = ggd.numPartialLogL( problem, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
#        print( "partalt = ", ged.partialLogL_alt( problem, lppar, lpfi ) )
        print( "numpart = ", nG )
        aG = ggd.partialLogL_alt( problem, ggpar, ggfi )
        print( "partalt = ", aG )

        assertAAE( dG, aG, 5 )
        assertAAE( dG[:3], dL, 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        ggscl = 0.1
        hypar = [ggscl, power]
        sigma = ggscl / ff
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [sigma] )
        logL = ged.logLikelihood( problem, lppar )
        lggL = ggd.logLikelihood( problem, ggpar )
        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        ggscl = 0.5
        hypar = [ggscl, power]
        sigma = ggscl / ff
        print( "Super : ", hypar, "  sigma : ", sigma )
        ggpar = numpy.append( param, hypar )
        lppar = numpy.append( param, [sigma] )

        problem.weights = self.wgt
        dL = ged.partialLogL( problem, lppar, lpfi )
        nL = ged.numPartialLogL( problem, lppar, lpfi )
        dG = ggd.partialLogL( problem, ggpar, ggfi)
        nG = ggd.numPartialLogL( problem, ggpar, ggfi )

        print( "partial = ", dL )
        print( "partial = ", dG )
        print( "numpart = ", nL )
#        print( "partalt = ", ged.partialLogL_alt( problem, lppar, lpfi ) )
        print( "numpart = ", nG )
        print( "partalt = ", ggd.partialLogL_alt( problem, ggpar, ggfi ) )
        assertAAE( dG[:3], dL, 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        for i in range( 11 ) :
            param = numpy.asarray( [i-5, 5, ggscl, power], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ggd.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def testExponentialErrorDistribution2( self ):

        print( "=======   Test Exponential Error Distribution 2  ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [0.9,10], dtype=float )
        ged = GaussErrorDistribution( )
        ggd = ExponentialErrorDistribution( )

        print( str( ged ) )
        print( str( ggd ) )

        self.assertTrue( ggd.acceptWeight() )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        data = [ -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 ]
        problem = ClassicProblem( model=poly, xdata=self.x, ydata=data )

        power = 2.0
        scale = 1.0
        hypar = [scale, power]
        sigma = ggd.toSigma( hypar )
        print( hypar, sigma )
        ff = 1 / sigma
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
#        gapar = numpy.append( param, [1.0] )

        chisq = ged.getChisq( problem, gapar )
        chipp = ggd.getChipow( problem, ggpar )
        print( "chipp = %8.3f   chisq = %8.3f"%( chipp, chisq ) )
#        assertAAE( chipp, chisq )


        logL = ged.logLikelihood( problem, gapar )
        lggL = ggd.logLikelihood( problem, ggpar )

        print( 'gapar  ', gapar )
        print( 'ggpar  ', ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        ggfi = numpy.array( [0,1,-2,-1] )
        lpfi = numpy.array( [0,1,-1] )

        dL = ged.partialLogL( problem, gapar, lpfi )
        nL = ged.numPartialLogL( problem, gapar, lpfi )
        dG = ggd.partialLogL( problem, ggpar, ggfi)
        nG = ggd.numPartialLogL( problem, ggpar, ggfi )

        print( "partial = ", dL )
        print( "numpart = ", nL )
        print( "partalt = ", ged.partialLogL_alt( problem, gapar, lpfi ) )
        print( "partial = ", dG )
        print( "numpart = ", nG )
        print( "partalt = ", ggd.partialLogL_alt( problem, ggpar, ggfi ) )

        print( nG[-2] * ff )

        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        scale = 0.1
        hypar = [scale,power]
        sigma = ggd.toSigma( hypar )
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
        logL = ged.logLikelihood( problem, gapar )
        lggL = ggd.logLikelihood( problem, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        dL = ged.partialLogL( problem, gapar, lpfi )
        nL = ged.numPartialLogL( problem, gapar, lpfi )
        dG = ggd.partialLogL( problem, ggpar, ggfi)
        nG = ggd.numPartialLogL( problem, ggpar, ggfi )

        print( "partial = ", dL )
        print( "numpart = ", nL )
        print( "partial = ", dG )
        print( "numpart = ", nG )
        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 4 )

        scale = 0.5
        hypar = [scale, power]
        sigma = ggd.toSigma( hypar )
        print( "Super : ", hypar, "  sigma : ", sigma )
        ggpar = numpy.append( param, hypar )
        gapar = numpy.append( param, [sigma] )
        problem.weights = self.wgt

        logL = ged.logLikelihood( problem, gapar )
        lggL = ggd.logLikelihood( problem, ggpar )

        print( "lggL  = %8.3f   logL  = %8.3f" % ( lggL, logL ) )
        assertAAE( logL, lggL )

        dL = ged.partialLogL( problem, gapar, lpfi )
        nL = ged.numPartialLogL( problem, gapar, lpfi )
        dG = ggd.partialLogL( problem, ggpar, ggfi)
        nG = ggd.numPartialLogL( problem, ggpar, ggfi )

        print( "partial = ", dL )
        print( "numpart = ", nL )
        print( "partial = ", dG )
        print( "numpart = ", nG )
        assertAAE( dG[:2], dL[:2], 5 )
        assertAAE( dG[2]*ff, dL[2], 5 )
        assertAAE( dG, nG, 5 )
        assertAAE( dL, nL, 5 )

        for i in range( 11 ) :
            param = numpy.asarray( [i-5,5,scale,power], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ggd.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def testModelDistribution( self ):
        print( "====== Test Model Distribution ======================" )
        poly = PolynomialModel( 1 )
        param = numpy.asarray( [12, 10], dtype=float )
        data = numpy.asarray( self.data + 13, dtype=int )
        print( "Data : ", data )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=data )

        ed = ModelDistribution( limits=[0.01,10] )
        printclass( ed )


    def testPoissonErrorDistribution( self ):
        print( "====== Test Poisson Error Distribution ======================" )
        poly = PolynomialModel( 1 )
        param = numpy.asarray( [12, 10], dtype=float )
        data = numpy.asarray( self.data + 13, dtype=int )
        print( "Data : ", data )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=data )

        ped = PoissonErrorDistribution( )
        print( str( ped ) )

        self.assertFalse( ped.acceptWeight() )

        logL = ped.logLikelihood( problem, param )
        print( "logL  = %8.3f"%( logL ) )
        scale = 0.1
        logL = ped.logLikelihood( problem, param )
        mok = problem.result( param )
        altL = ped.logLikelihood_alt( problem, param )

        lLdata = ped.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        print( "logL  =  %8.3f  ldata %8.3f  alt %8.3f" % (logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        logL = ped.logLikelihood( problem, [-5, 0] )
        print( "logL  = %8.3f"%( logL ) )
        self.assertTrue( math.isinf( logL ) )

        scale = 1.0
        fitIndex = [0,1]
        dL = ped.partialLogL( problem, param, fitIndex )
        nL = ped.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        scale = 0.5
        dL = ped.partialLogL( problem, param, fitIndex )
        nL = ped.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        aG = ped.partialLogL_alt( problem, param, fitIndex )
        print( "partalt = ", aG )
        assertAAE( dL, aG, 5 )

        scale = 1.0
        for i in range( 10 ) :
            param = numpy.asarray( [8+i,4], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ped.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )


    def testBernoulliErrorDistribution( self ):
        print( "====== Test Bernoulli Error Distribution ======================" )
        poly = LogisticModel( fixed={0:1} )
        param = numpy.asarray( [0.5,1.0], dtype=float )
        data = numpy.asarray( [0,0,0,0,0,1,1,1,1,1,1], dtype=int )
        data = numpy.asarray( [1,1,1,1,1,1,0,0,0,0,0], dtype=int )

        print( "x    : ", fmt( self.x, max=None ) )
        print( "Data : ", fmt( data, max=None ) )
        print( "yfit : ", fmt( poly.result( self.x, param ), max=None ) )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=data )

        ped = BernoulliErrorDistribution( )
        print( str( ped ) )

        self.assertFalse( ped.acceptWeight() )

        logL = ped.logLikelihood( problem, param )
        print( "logL  = %8.3f"%( logL ) )

        logL = ped.logLikelihood( problem, param )
        mok = problem.result( param )
        altL = ped.logLikelihood_alt( problem, param )

        lLdata = ped.logLdata( problem, param )
        print( "Ldata: ", fmt( lLdata, max=None ) )
        logL0 = numpy.sum( lLdata )
        print( "logL  =  %8.3f  ldata %8.3f  alt %8.3f" % (logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        print( "logL  = %8.3f  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        fitIndex = [0,1]
        dL = ped.partialLogL( problem, param, fitIndex )
        nL = ped.numPartialLogL( problem, param, fitIndex )
        aL = ped.partialLogL_alt( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "altpart = ", aL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

        print( "             ", fmt( [4.0/(k+1) for k in range( 9 )], max=None ) )
        for i in range( 10 ) :
            param = numpy.asarray( [i-4,1], dtype=float )
            print( fmt( param[0] ), "  :  ", end="" )
            for k in range( 9 ):
                param[1] = 4.0 / ( k + 1 )
                print( " %8.3f" % ped.logLikelihood( problem, param ), end="" )

            print( "" )


    def testCauchyErrorDistribution( self ):
        print( "\n   Test Cauchy Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        ced = CauchyErrorDistribution( )
        self.assertFalse( ced.acceptWeight() )

        param = numpy.asarray( [1,10,1], dtype=float )
        logL = ced.logLikelihood( problem, param )
        print( "logL  =  %8.3f" % ( logL ) )
        scale = 0.1
        s2 = scale * scale
        param[2] = scale
        logL = ced.logLikelihood( problem, param )
        altL = ced.logLikelihood_alt( problem, param )

        lLdata = ced.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        print( "logL  =  %8.3f  ldata %8.3f  alt %8.3f" % (logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

#        print( "logL  =  %8.3f  alt %8.3f" % (logL, altL ) )
#        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
        fi = [0,1,-1]
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 0.5
        param[2] = scale
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        aG = ced.partialLogL_alt( problem, param, fi )
        print( "partalt = ", aG )
        assertAAE( dL, aG, 5 )


        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, ":  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ced.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )


    def testLaplaceErrorDistribution( self ):
        print( "\n   Test Laplace Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        ced = LaplaceErrorDistribution( )
        print( str( ced ) )

        self.assertTrue( ced.acceptWeight() )

        param = numpy.asarray( [1,10,1], dtype=float )
        logL = ced.logLikelihood( problem, param )
        print( "logL  =  %8.3f" % ( logL ) )
        scale = 0.1
        param[2] = scale
        s2 = scale * scale
        logL = ced.logLikelihood( problem, param )

        altL = ced.logLikelihood_alt( problem, param )

        lLdata = ced.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        print( "logL  =  %8.3f  ldata %8.3f  alt %8.3f" % (logL, logL0, altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )


        scale = 1.0
        param[2] = scale
        fi = [0,1,-1]
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "partalt = ", ced.partialLogL_alt( problem, param, fi ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 0.5
        param[2] = scale
        problem.weights = self.wgt
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "partalt = ", ced.partialLogL_alt( problem, param, fi ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        logL = ced.logLikelihood( problem, param )
        cced = ced.copy()
        logLc = cced.logLikelihood( problem, param )
        print( cced )
        print( "logL  =  %8.3f  copy %8.3f" % (logL, logLc ) )
        dLc = cced.partialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        assertAAE( logL, logLc, 6 )
        assertAAE( dL, dLc, 6 )

        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, ":  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ced.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def testUniformErrorDistribution( self ):
        print( "\n   Test Uniform Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        ced = UniformErrorDistribution( )
        print( str( ced ) )

        print( "x   ", self.x )
        print( "y   ", self.data )
        self.assertTrue( ced.acceptWeight() )

        param = numpy.asarray( [0.3,11,4], dtype=float )
        print( "m   ", problem.result( param ) )
        logL = ced.logLikelihood( problem, param )
        altL = ced.logLikelihood_alt( problem, param )
        print( "logL  =  %8.3f  altL  %8.3f" % ( logL, altL ) )
        assertAAE( logL, altL )

        scale = 1.0
        param[2] = scale
        logL = ced.logLikelihood( problem, param )

        lLdata = ced.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )
        print( "logL  =  %8.3f  ldata %8.3f" % (logL, logL0 ) )
        assertAAE( logL, logL0 )

        print( "logL  =  %12.3g" % (logL) )
        self.assertTrue( logL == -math.inf )

        scale = 4.0
        param[2] = scale
        fi = [0,1,-1]
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "partalt = ", ced.partialLogL_alt( problem, param, fi ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 5
        param[2] = scale
        problem.weights = self.wgt
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        logL = ced.logLikelihood( problem, param )
        cced = ced.copy()
        logLc = cced.logLikelihood( problem, param )
        print( cced )
        print( "logL  =  %8.3f  copy %8.3f" % (logL, logLc ) )
        dLc = cced.partialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        assertAAE( logL, logLc, 6 )
        assertAAE( dL, dLc, 6 )

        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,10], dtype=float )
            print( param, ":  ", end="" )
            for k in range( 9 ):
                print( " %6.2f" % ced.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )

    def testMixedErrorDistribution0( self ):
        print( "\n   Test Mixed Error Distribution 0\n" )
        m = PolynomialModel( 0 )
        x = numpy.linspace( -10, 10, 201 )
        problem = ClassicProblem( m, x, x )
        ed1 = GaussErrorDistribution( )
#        ed2 = GaussErrorDistribution( )
        ed2 = UniformErrorDistribution( )

        ced = MixedErrorDistribution( ed1, ed2  )
        print( str( ced ) )

        param = numpy.asarray( [0.0, 0.3, 5, 0.7 ], dtype=float )
        if self.doplot :
            for k in range( 5 ) :
                param[3] = 0.25 * k
                lld = ced.logLdata( problem, param )
                plt.plot( x, numpy.exp( lld ) )
            plt.title( ced.__str__() )
            plt.show()



    def testMixedErrorDistribution( self ):
        print( "\n   Test Mixed Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        ed1 = GaussErrorDistribution( )
        ed2 = UniformErrorDistribution( )
        ced = MixedErrorDistribution( ed1, ed2  )

        self.assertTrue( ced.acceptWeight() )

        param = numpy.asarray( [0.3, 11, 1, 4, 0.7 ], dtype=float )

        logL = ced.logLikelihood( problem, param )
        print( "logL  =  %8.3f" % ( logL ) )
#        self.assertTrue( logL == -11 * math.log( 2 * param[2] ) )

        param[2:4] *= 0.097
#        param[4] = 0.7
        logL = ced.logLikelihood( problem, param )


        print( "logL  =  %8.3f" % (logL) )
        print( "lld  ", ced.logLdata( problem, param ) )
        pars = param[:3]
        print( "ll1  ", ced.errdis1.logLdata( problem, pars ) )
        pars[2] = param[3]
        print( "ll2  ", ced.errdis2.logLdata( problem, param ) )


        fi = [0,1,-3,-2,-1]
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

#        self.assertTrue( logL == -math.inf )

        scale = 2.0
        param[2] = 0.1 * scale      ## gauss
        param[3] = scale            ## uniform
        fi = [0,1,-3,-2,-1]
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        scale = 5
        param[2] = scale
        problem.weights = self.wgt
        dL = ced.partialLogL( problem, param, fi )
        nL = ced.numPartialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 2 )

        print( "make a copy" )
        logL = ced.logLikelihood( problem, param )
        cced = ced.copy()
        logLc = cced.logLikelihood( problem, param )
        print( cced )
        print( "logL  =  %8.3f  copy %8.3f" % (logL, logLc ) )
        dLc = cced.partialLogL( problem, param, fi )
        print( "params  = ", param, scale )
        print( "partial = ", dL )
        assertAAE( logL, logLc, 6 )
        assertAAE( dL, dLc, 6 )

        for i in range( 11 ):
            param = numpy.asarray( [i-5,5,0.2,1.0,0.7], dtype=float )
            print( param )
            for k in range( 9 ):
                print( " %10.3g" % ced.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( ErrorDistributionTest.__class__ )





import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .Model import Model
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2008 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class Sample( object ):
    """
    Sample is weighted random draw from a Posterior distribution as
    provided by a Sampler

    Each Sample maintains 5 attributes

    Attributes
    ----------
    id : int
        identification number
    parent : int
        id of the parent (-1 for Adam/Eve)
    model : Model
        the model being used
    logL : float
        log Likelihood = log Prob( data | params )
    logW : float
        log Weights of the log of the weight of the sample.
        The weight is the relative contribution to the evidence integral.
        logW = logL + log( width )
        The logZ, the evidence, equals the log of the sum of the contributions.
        logZ = log( sum( exp( logW ) ) )
    parameters : array_like
        parameters (of the model)
    nuisance : array_like (optional)
        nuisance parameters (of the problem)
    hyper : array_like (optional)
        list of hyper parameters (of the error distribution)
    fitIndex : array_like or None
        list of allpars to be fitted.
    allpars : array_like (read only)
        list of parameters, nuisance parameters and hyperparameters

    Author       Do Kester

    """

    def __init__( self, id, parent, start, model, parameters=None, fitIndex=None, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        id : int
            id of the sample
        parent : int
            id of the parent (-1 for Adam/Eve)
        start : int
            iteration in which the walker was constructed
        model : Model
            the model being used. Parameters are copied from this model.
        parameters : array_like
            list of model parameters
        fitIndex : array_like
            list of indices in allpars that need fitting
        copy : Sample
            the sample to be copied

        """
        self.id = id
        self.parent = parent
        self.start = start

        if copy is None :
            self.model = model
            self.parameters = parameters if parameters is not None else model.parameters
#            self.fitIndex = fitIndex if fitIndex is not None else numpy.arange( model.npars )
            self.fitIndex = fitIndex

            self.logL = 0.0
            self.logW = 0.0
        else :
            self.model = None if copy.model is None else copy.model.copy()
            self.fitIndex = None if copy.fitIndex is None else copy.fitIndex.copy()
#            self.model = copy.model.copy()
#            self.fitIndex = copy.fitIndex.copy()
            self.parameters = copy.parameters.copy()
            if hasattr( copy, "nuisance" ) : self.nuisance = copy.nuisance.copy()
            if hasattr( copy, "hyper" ) : self.hyper = copy.hyper.copy()
            self.logL = copy.logL
            self.logW = copy.logW

    def copy( self ):
        """
        Copy.

        """
        return Sample( self.id, self.parent, self.start, self.model, copy=self )

    def __getattr__( self, name ) :
        """
        Return the value of one of `parameters`, `scale`,

        """
        if name == "weight" :
            return math.exp( self.logW )
        elif name == "allpars" :
            allpars = self.parameters.copy()
            if hasattr( self, "nuisance" ) :
                allpars = numpy.append( allpars, self.nuisance )
            if hasattr( self, "hyper" ) :
                allpars = numpy.append( allpars, self.hyper )
            return allpars
        elif name == "hypars" :
            return self.hyper
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def __setattr__( self, name, value ) :
        """
        Set attributes.
        """
        if name == "parameters" :
            object.__setattr__( self, name, value )
            return

        key0 = [ "model", "fitIndex" ]
        key1 = {"id" : int, "parent" : int, "start" : int, "model": Model,
                "logL" : float, "logW" : float }
        key2 = {"fitIndex" : int, "nuisance" : float, "hyper" : float }
        if ( Tools.setNoneAttributes( self, name, value, key0 ) or
             Tools.setSingleAttributes( self, name, value, key1 ) or
             Tools.setListOfAttributes( self, name, value, key2 ) ) :
            pass
        else :
            raise AttributeError(
                "Object has no attribute " + name + " of type " + str( value.__class__ ) )

    def __str__( self ) :
        return str( "Sample: %3d parent: %3d model: %s logL: %10.2f logW: %10.2f"%
                    ( self.id, self.parent, self.model.shortName(), self.logL, self.logW ) )

    """
    def check( self, nhyp=0 ) :
#        print( len( self.allpars ), self.model.npchain, nhyp )
        if not len( self.allpars ) == ( self.model.npchain + nhyp ) :
            raise ValueError( "Sample inconsistent parameter length : %d is not ( %d + %d )" %
                ( len( self.allpars ), self.model.npchain, nhyp ) )
        if nhyp > 0 and self.allpars[-nhyp] <= 0 :
            raise ValueError( "Sample has non-positive hyperparameter: %f" % self.allpars[-nhyp] )
    """



import numpy as numpy
import math
import warnings
import matplotlib.pyplot as plt
from astropy.table import Table

from .ImageAssistant import ImageAssistant
from .MonteCarlo import MonteCarlo
from .ConvergenceError import ConvergenceError
from . import Tools
from . import Plotter
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2023 Do Kester

class BaseFitter( object ):
    """
    Base class for all Fitters.

    The Fitter class is to be used in conjunction with *Model classes.

    The Fitter class and its descendants fit data to a model. Fitter itself
    is the variant for linear models, ie. models linear in its parameters.

    For both linear and nonlinear models it holds that once the optimal
    estimate of the parameters is found, a variety of calculations is exactly
    the same: standard deviations, noise scale, evidence and model errors.
    They all derive more or less from the inverse Hessian matrix ( aka the
    covariance matrix ). All these calculations are in this Fitter class.
    Other Fitter classes relegate their calculation in these issues to this one.

    Examples
    --------
    # It is not possible to use this class. User Fitter, CurveFitter etc. in stead

    Note Also
    ---------
    1. The calculation of the evidence is an Gaussian approximation which is
       only exact for linear models with a fixed scale.
    2. Attributes labelled as read only should not be set by a user.

    Author: Do Kester

    Attributes
    ----------
    model : Model
        the model to be fitted
    xdata : array_like
        independent variable(s)
    nxdata : int (read only)
        length of the xdata vector(s)
    ndim : int (read only)
        number of xdata vectors
    weights : array_like
        the weights on the data from the last fit
    imageAssistant : ImageAssistant
        to convert images to pixels indices, needed for a fit
    keep : dict of {int : float}
        to keep the indexed (int) parameter at the provided value (float)
    fitIndex : list of int (or None)
        list of parameter indices to fit (None is all)
    npfit : int
        the number of parameters fitted in the last fit.
    fixedScale : float
        the fixed noise scale.
        The presence of `fixedScale` has consequences for the definitions of `chisq`,
        `(co)variance`, `stdevs` and `evidence`

    minimumScale : float
        introduce a minimum value for the noise scale
    design : matrix (read only)
        the design matrix (partial of model to parameters)
        returns self.getDesign()

    Attributes (available after a call to fit())
    ----------
    yfit : array_like
        The model result at the optimal value for the parameters.
        If map is true, a map is returned.
    chisq : float (read only)
        chisquared of the fit
    parameters : ndarray
        parameters fitted to the model
        returns self.model.parameters
    stdevs, standardDeviations : array_like (read only)
        the standard deviations on the parameters
        returns self.getStandardDeviations()
    hessian : matrix (read only)
        the hessian matrix
    covariance : matrix (read only)
        the covariance matrix
        returns self.getCovarianceMatrix()
    scale : float
        the noise scale
        returns self.getScale()
    sumwgt : float (read only)
        sum of the weights
    logZ : float (read only)
        the e-log of the evidence
        returns self.getLogZ()
    evidence : float (read only)
        the 10log of the evidence (logZ / log(10))
        returns self.getEvidence()

    Attributes (available after a call to getLogZ() or getEvidence())
    ----------
    logOccam : float (read only)
        Occam factor
    logLikelihood : float (read only)
        log of the likelihood

    """

    #  *****CONSTRUCTORS********************************************************
    def __init__( self, xdata, model, map=False, keep=None, fixedScale=None ):
        """
        Create a new Fitter, providing inputs and model.

        A Fitter class is defined by its model and the input vectors ( the
        independent variable ). When a fit to another model and/or another
        input vector is needed a new object should be created.

        Parameters
        ----------
        xdata : array_like
            independent input variable(s)
        model : Model
            the model function to be fitted
        map : bool (False)
            When true, the xdata should be interpreted as a map.
            The fitting is done on the pixel indices of the map,
            using ImageAssistant
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep will be used by the Fitter as long as the Fitter exists.
            See also fit( ..., keep=dict )
        fixedScale : None or float
            None : the noise scale is not fixed
            float: value of the fixed noise scale
            The value of fixedScale only influences the evidence calculation

        Raises
        ------
        ValueError when one of the following is true
            1. Dimensionality of model and input does not match.
            2. Nans in input stream.
            3. Model is not the head of a compound model chain.

        """
        if model != model._head:
            raise ValueError( "Model is not the head of a compound model chain" )

        if map :
            self.imageAssistant = ImageAssistant()
            self.xdata = self.imageAssistant.getIndices( xdata )
        else :
            self.imageAssistant = None
            self.xdata = numpy.asarray( xdata )

        if isinstance( xdata, Table ) :
            ndim = len( xdata.columns )
        else :
            if numpy.any( numpy.isnan( xdata ) ) :
                raise ValueError( "NaNs in xdata array" )

            ndim = 1 if self.xdata.ndim <= 1 else self.xdata.shape[1]

        ninp = Tools.length( xdata )
        self.nxdata = ninp
        self.ndim = ndim
        self.model = model
        self.keep = keep
        self.fitIndex = self.keepFixed( keep )
        self.fixedScale = fixedScale

        if self.ndim != model.ndim:
            raise ValueError( "Model (%d) and xdata (%d) must be of the same dimensionality."
                                % (model.ndim, self.ndim) )

    def setMinimumScale( self, scale=0 ) :
        """
        Introduce a minimum in scale calculation and consequently in chisq.
            chi^2 >= sumwgt * scale^2

        Parameters
        ----------
        scale : float
            minimum scale
        """
        self.minimumScale = scale


    def fitprolog( self, ydata, weights=None, accuracy=None, keep=None ) :
        """
        Prolog for all Fitters.

        1. Checks data/weighs/accuracy for Nans
        2. Makes fitIndex.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
        accuracy : float or array_like
            accuracy of (individual) data
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)

        Returns
        -------
        fitIndex : ndarray of int
            Indices of the parameters that need fitting
        ydata : ndarray
            Only different from input when ydata is a map 
        fitwgts : float or ndarray
            Combines weights and accuracy into ( weights / accuracy^2 )
            1.0 if both are None
        """
        self.checkNan( ydata, weights=weights, accuracy=accuracy )

        if self.imageAssistant is not None :
            ydata = self.imageAssistant.getydata( ydata )
            if weights is not None :
                weights = self.imageAssistant.getydata( weights )
            if Tools.length( accuracy ) > 1  :
                accuracy = self.imageAssistant.getydata( accuracy )
        else :
            ydata = numpy.asarray( ydata )
            if weights is not None :
                weights = numpy.asarray( weights )

        self.weights = weights
        self.accuracy = accuracy

        self.fitWgts = 1.0
        self.sumwgt = self.nxdata
        if weights is not None : 
            self.fitWgts *= weights
            self.sumwgt = numpy.sum( weights )
        if accuracy is not None :
            self.fitWgts /= ( accuracy * accuracy )
           
        if keep is not None :
            return ( self.keepFixed( keep ), ydata, self.fitWgts )

        if self.fitIndex is None :
            self.npfit = self.model.npchain
            return ( numpy.arange( self.model.npchain, dtype=int ), 
                     ydata, self.fitWgts )

        self.npfit = len( self.fitIndex )
        return ( self.fitIndex, ydata, self.fitWgts )

    def fitpostscript( self, ydata, plot=False ) :
        """
        Produce a plot of the results.
        """
        if plot :
            self.plotResult( xdata=self.xdata, ydata=ydata, model=self.model,
                    residuals=True, confidence=False, show=True )

# *****KEEP**************************************************************
    def keepFixed( self, keep=None ) :
        """
        Keeps parameters fixed at the provided values.

        1. The model will act exactly as if it were a model with less
           parameters, although slightly less efficient.
        2. Repeated calls start from scratch.
        3. Reset with keepFixed( None )

        Parameters
        ----------
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)

        Returns
        -------
        fitIndex : list of int (or None)
            list of parameter indices to be kept
        """
        if keep is None :
            self.npfit = self.model.npchain
            return None
        else :
#            print( "Keep  ", keep )
            self.npfit = len( keep )
            fitIndex = numpy.arange( self.model.npchain )      # start from scratch
            fitIndex = numpy.setxor1d( fitIndex, list( keep.keys() ) )
            self.model.parameters[list(keep.keys())] = list( keep.values() )
            return fitIndex

    def insertParameters( self, fitpar, index=None, into=None ) :
        """
        Insert fitparameters into the parameters when fitIndex is present.

        Parameters
        ----------
        fitpar : list of float
            (fitted) parameters
        index : list of int
            list of parameter indices to be kept
        into : list of float
            array into which the fitpar need to be inserted.

        """
        if index is None :
            index = self.fitIndex

        fitpar = numpy.array( fitpar, dtype=float, copy=False, ndmin=1 )

        if index is None or len( fitpar ) == len( self.model.parameters ) :
            return fitpar
        pars = self.model.parameters.copy() if into is None else into
        pars[index] = fitpar
        return pars

    #  *****FIT*****************************************************************
    def modelFit( self, ydata, weights=None, keep=None ):
        """
        Return model fitted to the data.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted.
        weights : None or array_like
            weights to be used
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.

        """
        self.model.parameters = self.fit( ydata, weights=weights, keep=keep )
        return self.yfit


    def limitsFit( self, fitmethod, ydata, weights=None, keep=None ) :
        """
        Fit the data to the model.
        When a parameter(s) transgresses the limits, it set and fixed at that limit
        and the fit is done again, excluding the parameter(s)
        When the chisq landscape is largely monomodal (no local minima) this is OK.

        Parameter
        ---------
        fitmethod : callable fitmethod( ydata, weights=weights )
            A fit method from the BaseFitter family
        ydata : array_like
            data that the model needs to be fit to
        weights : array_like
            weights partaining to the data.
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.

        Returns
        -------
        pars : array_like
            the parameters of the fit

        Raises
        ------
        Warning when parameters have been reset at the limits.

        """
        pars = self.fit( ydata, weights=weights, keep=keep )          # perform the fit
#        pars = fitmethod( ydata, weights=weights )          # perform the fit
        if self.model.priors is None :                  # no priors -> no limits
            return pars

        sfix = self.keep                                # save original fixed setting

        npchain = self.model.npchain
        params = self.model.parameters
        index = self.fitIndex
#        print( "index  ", fmt( index ) )
        if index is None :
            index = [*range( npchain )]                 # unpack range object into list
            fix = {}
        else :
            fix = sfix.copy()

        if keep is not None :
            for k in keep.keys() : 
                fix[k] = keep[k]

#        print( "index  ", fmt( index ) )
#        print( "fix    ", fmt( fix ) )

        ool = 0
        for k in index :                                    # check limits for params
            if params[k] < self.model.getPrior( k ).lowLimit :
                params[k] = self.model.getPrior( k ).lowLimit
                fix[k] = params[k]
                ool += 1
            elif params[k] > self.model.getPrior( k ).highLimit :
                params[k] = self.model.getPrior( k ).highLimit
                fix[k] = params[k]
                ool += 1

        if ool > 0 :                                        # some transgressions
            print( "ool    ", fmt( ool ), fmt( fix ) )
            self.keepFixed( fix )                           # fix them
#            pars = fitmethod( ydata, weights=weights )      # run fit again
            pars = self.fit( ydata, weights=weights, keep=fix )      # run fit again
#            print( "pars   ", fmt( pars, max=None ) )
#            print( fmt( sfix ) )
            self.keepFixed( sfix )                          # reset original fitIndex
            if sfix is not None :                           # select parameters
                pars = [self.model.parameters[k] for k in self.fitIndex]
                fix = list( set( fix ) - set( sfix ) )
#            warnings.warn( "Parameters ", fix, " exceeded limits." )

        return pars                                         # return parameters


    def fit( self, ydata, weights=None, keep=None ) :
        """
        Return model parameters fitted to the data.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted.
        weights : array_like
            weights to be used
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.

        Raises
        ------
        NotImplementedError. BaseFitter cannot perform fits by itself.

        """
        raise NotImplementedError( "BaseFitter is a base class, not suitable itself to perform fits." )

    def checkNan( self, ydata, weights=None, accuracy=None ):
        """
        Check there are no Nans or Infs in ydata or weights or accuracy.
        Check also for zeros or negatives in accuracy.

        Parameters
        ----------
        ydata : array_like
            data to be fitted.
        weights : array_like
            weights pertaining to ydata
        accuracy : float or array_like
            accuracy of (individual) data

        Raises
        ------
        ValueError.
        """
        if not numpy.all( numpy.isfinite( ydata ) ) :
            raise ValueError( "Fitter: NaNs or Infs in ydata" )
        if  weights is not None and not numpy.all( numpy.isfinite( weights ) ) :
            raise ValueError( "Fitter: NaNs or Infs in weights" )
        if  accuracy is not None :
            if not numpy.all( numpy.isfinite( accuracy ) ) :
                raise ValueError( "Fitter: NaNs or Infs in accuracy" )
            if not numpy.all( accuracy > 0 ) :
                raise ValueError( "Fitter: zeros or negatives in accuracy" )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name in ['logOccam', 'logLikelihood', 'chisq'] :
            raise AttributeError( str( self ) + ": " + name + " is not yet available." )
        elif name == 'parameters' :
            return self.model.parameters
        elif name == 'weights' :            ## not present return None
            return None
        elif name == 'fitWgts' :            ## not present return None
            return None
        elif name == 'sumwgt' :             ## not present return nxdata
            return self.nxdata
        elif name == 'yfit' :
            yfit = self.model.result( self.xdata )
            return yfit if self.imageAssistant is None else self.imageAssistant.resizeData( yfit )
        elif name == 'design' :
            return self.getDesign()
        elif name == 'hessian' :
            return self.getHessian()
        elif name == 'covariance' :
            return self.getCovarianceMatrix()
        elif name == 'stdevs' or name == 'standardDeviations' :
            return self.getStandardDeviations()
        elif name == 'scale' :
            return self.getScale()
        elif name == "stdevScale" :
            return self.scale / math.sqrt( 2 * self.sumwgt )
        elif name == 'evidence' :
            return self.getEvidence()
        elif name == 'logZ' :
            return self.getLogZ()
        else :
            raise AttributeError( str( self ) + ": Unknown attribute " + name )

        return None


    #  *****VECTOR**************************************************************
    def getVector( self, ydata, index=None ):
        """
        Return the &beta;-vector.

        It includes "normalized" data if present. See normalize().

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted. When such is appliccable, it should be
            multiplied by weights and/or appended by normdata.
        index : list of int
            index of parameters to be fixed

        """
        if self.model.isNullModel() :
            return numpy.asarray( 0 )
        design = self.getDesign( index=index )

        return numpy.inner( design.transpose(), ydata )

    #  *****HESSIAN**************************************************************
    def getHessian( self, params=None, weights=None, index=None ):
        """
        Calculates the hessian matrix for a given set of model parameters.

        It includes "normalized" data if present. See normalize()

        Parameters
        ----------
        params : array_like
            the model parameters to be considered
        weights : array_like
            weights to be used
        index : list of int
            index of parameters to be fixed

        """
        if params is None : params = self.model.parameters
        if weights is None : weights = self.fitWgts

#        self.sumwgt = self.nxdata if weights is None else numpy.sum( weights )

        if self.model.isNullModel() :
            return

        design = self.getDesign( xdata=self.xdata, params=params, index=index )

        if hasattr( self, "normweight" ) :
            if weights is None :
                weights = numpy.ones( self.nxdata, dtype=float )
            elif Tools.isInstance( weights, float ) :
                weights = numpy.full( self.nxdata, weights, dtype=float )
            weights = numpy.append( weights, self.normweight )

        design = design.transpose()

        if weights is not None :
            self.hessian = numpy.inner( design, design * weights )
        else :
            self.hessian = numpy.inner( design, design )

        return self.hessian

#      * TBD Condition number see Wikipedia: Condition Number and Matrix Norm

    #  *************************************************************************
    def getInverseHessian( self, params=None, weights=None, index=None ):
        """
        Return the inverse of the Hessian Matrix, H.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted.
        weights : array_like
            weights to be used
        index : list of int
            index of parameters to be fixed

        """
        hes = self.getHessian( params, weights, index )
        inh = numpy.linalg.inv( hes )
        return inh

    def getCovarianceMatrix( self ):
        """
        Returns the inverse hessian matrix over the fitted parameters,
                multiplied by the variance.

        Stdevs are found from this as np.sqrt( np.diag( covarianceMatrix ) )

        """
        cov =  self.getInverseHessian( index=self.fitIndex ) * self.makeVariance()
        return cov

    #  *************************************************************************
    def makeVariance( self, scale=None ):
        """
        Return the (calculated) variance of the remaining noise. I.e.
            var = chisq / dof
        when automatic noise scaling is requested or
            var = scale^2
        when we have a fixed scale.

        Parameters
        ----------
        scale : float
            noise scale to be used
        """
        if scale is not None :
            return scale * scale

        scale = self.getScale()
        var = scale * scale
        if hasattr( self, "minimumScale" ) :       # add minimum to scale when requested
            var += self.minimumScale * self.minimumScale
        return var

    def normalize( self, normdfdp, normdata, weight=1.0 ) :
        """
        If for some reason the model is degenerate, e.g when two parameters measure
        essentially the same thing, This method can disambiguate these parameters.

        It is like adding a dummy measurement of one (or more) parameter to the data.

        Parameters
        ----------
        normdfdp : array_like
            for each parameter to sum to value (same length as self.parameters)
        normdata : float
            simulated data value
        weight : float
            weight of this measurement
        """
        normdfdp = numpy.asarray( normdfdp, dtype=float )
        if normdfdp.ndim == 1 :
            normdfdp = normdfdp.reshape( 1, len( normdfdp ) )
            normdata = numpy.asarray( [normdata], dtype=float )
            weight   = numpy.asarray( [weight], dtype=float )

        if not hasattr( self, "normdfdp" ) :
            self.normdfdp = normdfdp
            self.normdata = normdata
            self.normweight = weight
        else :
            self.normdfdp = numpy.append( self.normdfdp, normdfdp, axis=0 )
            self.normdata = numpy.append( self.normdata, normdata )
            self.normweight = numpy.append( self.normweight, weight )


    #  *****DESIGN**************************************************************
    def getDesign( self, params=None, xdata=None, index=None ):
        """
        Return the design matrix, D.
        The design matrix is also known as the Jacobian Matrix.

        Parameters
        ----------
        xdata : array_like
            the independent input data
        params : array_like
            parameters of the model
        index : list of int
            index of parameters to be fixed

        """
        if params is None : params = self.model.parameters
        if xdata is None :  xdata = self.xdata

        design = self.model.partial( xdata, params )
        if hasattr( self, "normdfdp" ) :
            design = numpy.append( design, self.normdfdp, axis=0 )

        if index is not None :
            design = design[:,index]

        return design

    #  *****CHI-SQUARED*********************************************************
    def chiSquared( self, ydata, params=None, weights=None ):
        """
        Calculates Chi-Squared for data and weights.

        It is the (weighted) sum of the squared residuals.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted.
        params : array_like
            parameters for the model
        weights : array_like
            weights to be used

        Raises
        ------
        ValueError when chisq <= 0.

        """
        res2 = numpy.square( ydata - self.model.result( self.xdata, params ) )
        if weights is not None:         ## for weight and accuracy
            res2 *= weights

        self.chisq = numpy.sum( res2 )
        if self.chisq <= 0 :
            raise ValueError( str( self ) + ": chisq <= 0" )

        return self.chisq

    #  *****STANDARD DEVIATION**************************************************
    def getStandardDeviations( self ):
        """
        Calculates of standard deviations pertaining to the parameters.

            &sigma;_i = s * sqrt( C[i,i] )

        where C is the Covariance matrix, the inverse of the Hessian Matrix and
        s is the noiseScale.

        Standard deviation are calculated for the fitted parameters only.

        Note that the stdev will decrease with sqrt( N ) of the number of
        datapoints while the noise scale, s, does not.

        """
        cov = self.covariance

        stdevs = numpy.sqrt( cov.diagonal() )
        if self.keep is None :
            self.model.stdevs = stdevs
        else :
            intostd = numpy.zeros( self.model.npchain, dtype=float )
            self.model.stdevs = self.insertParameters( stdevs, into=intostd )

        return self.model.stdevs

    #  *****MONTE CARLO ERROR***************************************************
    def monteCarloError( self, xdata=None, monteCarlo=None):
        """
        Calculates &sigma;-confidence regions on the model given some inputs.

        From the full covariance matrix (inverse of the Hessian) random
        samples are drawn, which are added to the parameters. With this new
        set of parameters the model is calculated. This procedure is done
        by default, 25 times.
        The standard deviation of the models is returned as the error bar.

        The calculation of the confidence region is delegated to the class
        MonteCarlo. For tweaking of that class can be done outside BaseFitter.

        Parameters
        ----------
        xdata : array_like
            input data over which to calculate the error bars.
        monteCarlo : MonteCarlo
            a ready-made MonteCarlo class.

        """
        if xdata is None : xdata = self.xdata
        if monteCarlo is None :
            monteCarlo = MonteCarlo( xdata, self.model, self.covariance,
                                          index=self.fitIndex )

        return monteCarlo.getError( xdata )

    #  *************************************************************************
    def getScale( self ):
        """
        Return
        ------
        float : the noise scale
            scale = sqrt( chisq / DOF )

        Raise
        -----
        RuntimeError when DoF <= 0. The number of (weighted) datapoints is too small.

        """
        dof = self.sumwgt - self.npfit
        if dof <= 0 :
            raise RuntimeError( "More parameters than (weighted) data points" )

        return math.sqrt( self.chisq / dof )

    #  *****EVIDENCE************************************************************
    def getEvidence( self, limits=None, noiseLimits=None ):
        """
        Calculation of the evidence, log10( Z ), for the model given the data.

            E = log10( P( Model | data ) )

        The calculation of the evidence uses a Gaussion approximation of the Posterior
        probability.
        It needs to know the limits of the parameters (and the noise scale if applicable),
        either from the priors in the model or from keywords "limits/noiseLimits".


        Parameters
        ----------
        limits : list of 2 floats/array_likes
            possible range of the parameters. ( [low,high] )
        noiseLimits : list of 2 floats
            possible range on noise scale ( [low,high] )

        Raises
        ------
        ValueError when no Prior is available

        """
        return self.getLogZ( limits, noiseLimits ) / math.log( 10.0 )

    def getLogLikelihood( self, autoscale=False, var=1.0 ) :
        """
        Return the log likelihood.

        It is implementing eq 19/20 last parts (Kester 2002) term by term

        Parameters
        ----------
        autoscale : bool
            whether the noise scale is optimized too
        var : float
            variance
        """

        if self.accuracy is not None :
            if Tools.length( self.accuracy ) > 1 :
                slacc = numpy.sum( numpy.log( self.accuracy ) )
            else :
                slacc = self.sumwgt * numpy.log( self.accuracy )
        else :
            slacc = 0

#        print( "BF1    ", var, slacc, self.chisq, self.sumwgt )

        return -0.5 * ( self.sumwgt * math.log( 2 * math.pi * var ) +
                        self.chisq / var ) - slacc
            
    def getLogZ( self, limits=None, noiseLimits=None ):
        """
        Calculation of the evidence, log( Z ), for the model given the data.

            logZ = log( P( Model | data ) )

        The calculation of the evidence uses a Gaussion approximation of the Posterior
        probability.
        It needs to know the limits of the parameters (and the noise scale if applicable),
        either from the priors in the model or from keywords "limits/noiseLimits".


        Parameters
        ----------
        limits : list of 2 floats/array_likes
            possible range of the parameters. ( [low,high] )
        noiseLimits : list of 2 floats
            possible range on noise scale ( [low,high] )

        Raises
        ------
        ValueError when no Prior is available

        RuntimeError when DoF <= 0. The number of (weighted) datapoints is too small.

        """
        dof = self.sumwgt - self.npfit
        if dof <= 0 :
            raise RuntimeError( "More parameters than (weighted) data points" )

        if noiseLimits is not None :
            scalerange = math.log( noiseLimits[1] ) - math.log( noiseLimits[0] )
            autoScale = True
            s2 = self.makeVariance( )
            self.logOccam = 0.5 * math.log( math.pi * dof ) - scalerange
        else :
            autoScale = False
            scale = 1.0 if self.fixedScale is None else self.fixedScale
#            print( "BF    ", self.fixedScale, scale )
            s2 = scale * scale
            self.logOccam = 0.0

        # add integral over the scale prior, if scale is to be fitted.
        spr = math.log( scalerange ) if autoScale else 0.0

        # obtain the loglikelihood.
        self.logLikelihood = self.getLogLikelihood( autoscale=autoScale, var=s2 )
#        print( "BF2    ", self.logLikelihood )


        if self.npfit == 0 :
            return self.logLikelihood + self.logOccam

        priors = self.model.priors
        if limits is not None :
            prirange = Tools.toArray( limits[1] ) - Tools.toArray( limits[0] )
        elif self.model.hasLimits() :
            prirange = [p.getIntegral() for p in priors]        # convert to list
        elif self.model.npchain > 0 :
            raise ValueError( "No limits provided on the parameters. Cannot calculate evidence." )

        priorlength = Tools.length( prirange )                         # maybe less than npar
        prirange = numpy.log( numpy.asarray( prirange ) )

        if priorlength == 1 :
            spr += prirange[0]
        elif limits is None and self.fitIndex is not None :
            fi = self.fitIndex
            q = numpy.where( fi < priorlength )
            fq = fi[q]
            priorlength = len( fq )
            spr += numpy.sum( prirange[fq] )
        else :
            spr += numpy.sum( prirange )

        if priorlength < self.npfit :                          # add enough times the last one
            spr += ( self.npfit - priorlength ) * prirange[-1]

        lidet = math.log( numpy.linalg.det( self.hessian ) )

        # implementing eq 18 (Kester 2002) term by term
        self.logOccam += -spr + 0.5 * ( self.npfit *
                         math.log( 2 * math.pi * s2 ) - lidet )

        return self.logLikelihood + self.logOccam

    def __str__( self ):
        """ Return name of the fitter.  """
        return "BaseFitter"


    def plotResult( self, xdata=None, ydata=None, model=None, residuals=True,
                    confidence=False, show=True ) :
        """
        Plot the results of the fit.

        Parameters
        ----------
        xdata : array_like
            xdata of the problem
        ydata : array_like
            ydata of the problem
        model : Model
            the model the ydata are fitted to at xdata.
        residuals : bool
            plot the residuals in a separate panel.
        confidence : bool
            plot confidence region
        show : bool
            display the plot.
        """
        if xdata is None :
            xdata = self.xdata
        if model is None :
            model = self.model
        fitter = self if confidence else None

        Plotter.plotFit( xdata, data=ydata, model=model, show=show,
                fitter=fitter, residuals=residuals )





import numpy as numpy
from scipy import special
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .HyperParameter import HyperParameter
from .NoiseScale import NoiseScale

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2023 Do Kester


class ExponentialErrorDistribution( ScaledErrorDistribution ):
    """
    Also known as generalized gaussian errordistribution.

    To calculate an Exponential likelihood.

    For one residual, x, it holds

        f( x ) = p / ( 2 s &Gamma;( 1 / p ) ) exp( - ( |x| / s ) ^ p )

    where s is the scale and p is the power.
    s and p are hyperparameters, which might be estimated from the data.

    The variance of this function is

        &sigma; ^ 2 = s ^ 2 &Gamma;( 3 / p ) / &Gamma;( 1 / p )

    See toSigma()

    The function is mostly used to calculate the likelihood L over N residuals,
    or easier to use log( L )

        logL = log( N p / ( 2 s &Gamma;( 1 / p ) ) ) - &sum;( ( |x| / s ) ^ p )

    Using weights this becomes:

        logL = log( &sum;( w ) p / ( 2 s &Gamma;( 1 / p ) ) ) - &sum;( w ( |x| / s ) ^ p )

    Note
    ----
    The scale s in Exponential is NOT the same as the scale in Gaussian or in Laplace.

    Attributes from ErrorDistibution
    --------------------------------
    hyperpar, deltaP, ncalls, nparts, sumweight, ndata, hypar, nphypar


    Author       Do Kester.

    """
    LOG2PI = math.log( 2 * math.pi )
    PARNAMES = ["scale", "power"]

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, power=2.0, limits=None, copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        scale : float
            noise scale
        power : float
            power of the distribution
        limits : None or [low,high] or [[low],[high]]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            [low]   low limit on [scale,power] (need to be >0)
            [high]  high limit on [scale,power]
            when limits are set, the scale cq. power are *not* fixed.
        copy : ExponentialErrorDistribution
            distribution to be copied.
        """
        super( ExponentialErrorDistribution, self ).__init__( limits=None, copy=copy )

        plim = None
        if limits is None :
            slim = None
        else :
            lo = limits[0]
            hi = limits[1]
            try :
                slim = [lo[0],hi[0]]
            except :
                slim = [lo,hi]
            try :
                plim = [lo[1],hi[1]]
            except :
                pass

        if copy is None :
            self.hyperpar = [NoiseScale( scale=scale, limits=slim ),
                             HyperParameter( hypar=power, limits=plim )]
        else :
            self.hyperpar = copy.hyperpar

    def copy( self ):
        """ Return copy of this.  """
        return ExponentialErrorDistribution( copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True

    def toSigma( self, hypar ) :
        """
        Return sigma, the squareroot of the variance.
        Parameter
        --------
        hypar : array_like (2 floats)
            the [scale,power] of this Exponential distribution.
        """
        p = hypar[1]
        return hypar[0] * math.sqrt( special.gamma( 3.0 / p ) / special.gamma( 1.0 / p ) )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem

        """
        self.ncalls += 1

        scale = allpars[-2] if not problem.hasAccuracy else problem.accuracy
        power = allpars[-1]

        chipow = self.getChipow( problem, allpars, scale )

        if isinstance( scale, float ) :
            norm = problem.sumweight * math.log( 0.5 * power / scale )
        elif problem.hasWeights() :
            norm = numpy.sum( problem.weights * numpy.log( 0.5 * power / scale ) )
        else :
            norm = numpy.sum( numpy.log( 0.5 * power / scale ) )

        sumpow = problem.sumweight * special.gammaln( 1.0 / power )

        return norm - sumpow - chipow

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        scale = allpars[-2] if not problem.hasAccuracy else problem.accuracy
        power = allpars[-1]

        res = problem.residuals( allpars[:-2], mockdata=mockdata )
        lld = - numpy.power( numpy.abs( res / scale ), power )

        norm = numpy.log( power / ( 2 * scale ) ) - special.gammaln( 1.0 / power )
        lld += norm
        if problem.weights is not None :
            lld *= problem.weights
        return lld

    def getChipow( self, problem, allpars=None, scale=1 ) :
        """
        Return chisq.

        return Sum over the (weighted) powered residuals

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        scale : float or array_like
            present scale
        """
        res = problem.residuals( allpars[:-2] ) / scale
        power = allpars[-1]

        ares = numpy.power( numpy.abs( res ), power )
        if problem.weights is not None :
            ares = ares * problem.weights
        return numpy.sum( ares )

    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale calculated from the residuals.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        power = allpars[-1]
        chi = self.getChipow( problem, allpars=allpars )
        return math.pow( chi / problem.sumweight, 1.0 / power )


    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        dL/ds is not implemented for problems with accuracy
        
        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.ncalls += 1

        scale = allpars[-2] if not problem.hasAccuracy else problem.accuracy
        power = allpars[-1]
        res = problem.residuals( allpars[:-2] )

        ars = numpy.abs( res / scale )
        rsp = numpy.power( ars, power )
        if problem.weights is not None :
            rsp = rsp * problem.weights

        dLdm = power * rsp / res
        dM = problem.partial( allpars[:-2] )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for  k in fitIndex :
            if k >= 0 :
                dL[i] = numpy.sum( dLdm * dM[:,k] )
                i += 1
            elif k == -2 :
                dL[-2] = - problem.sumweight / scale + power * numpy.sum( rsp ) / scale
            else :
                # special.psi( x ) is the same as special.polygamma( 1, x )
                dldp = problem.sumweight * ( power + special.psi( 1.0 / power ) )
                dldp /= ( power * power )
                dldp -= ( numpy.sum( rsp * numpy.log( ars ) ) )
                dL[-1] = dldp

        return dL

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative of all elements of the log( likelihood )
        to the parameters.

        dL/ds is not implemented for problems with accuracy
        
        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated by the model

        """
        param = allpars[:-2]
        res = problem.residuals( param, mockdata=mockdata )
        scale = allpars[-2] if not problem.hasAccuracy else problem.accuracy
        power = allpars[-1]

        ars = numpy.abs( res / scale )
        rsp = numpy.power( ars, power )
        if problem.weights is not None :
            rsp = rsp * problem.weights
            wgt = problem.weights
        else :
            wgt = 1.0

        dLdm = power * rsp / res
        dM = problem.partial( param )
##      TBD import mockdata into partial
#        dM = problem.partial( param, mockdata=mockdata )

        # special.psi( x ) is the same as special.polygamma( 1, x )
        dlp = wgt * ( power + special.psi( 1.0 / power ) ) / ( power * power )

        for  k in fitIndex :
            if k >= 0 :
                yield ( dLdm * dM[:,k] )
            elif k == -2 :
                yield ( power * rsp - wgt ) / scale
            else :
                yield dlp - rsp * numpy.log( ars )

#        return

    def __str__( self ) :
        return "Exponential error distribution"

